# #############################################################################
# trainable_recon.py
# ==================
# Authors :
# Yohann PERRON [yohann.perron@gmail.com]
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################

import pathlib as plib
from matplotlib import pyplot as plt
from lensless.recon.recon import ReconstructionAlgorithm
from lensless.utils.plot import plot_image
from lensless.recon.rfft_convolve import RealFFTConvolve2D

try:
    import torch

except ImportError:
    raise ImportError("Pytorch is require to use trainable reconstruction algorithms.")


class TrainableReconstructionAlgorithm(ReconstructionAlgorithm, torch.nn.Module):
    """
    Abstract class for defining lensless imaging reconstruction algorithms with trainable parameters.

    The following abstract methods need to be defined:

    * ``_update``: updating state variables at each iterations.
    * ``reset``: reset state variables.
    * ``_form_image``: any pre-processing that needs to be done in order to view the image estimate, e.g. reshaping or clipping.

    One advantage of deriving from this abstract class is that functionality for
    iterating, saving, and visualization is already implemented, namely in the
    ``apply`` method.

    Consequently, using a reconstruction algorithm that derives from it boils down
    to four steps:

    1. Creating an instance of the reconstruction algorithm.
    2. Training the algorithm
    3. Setting the data.
    4. Applying the algorithm.


    """

    def __init__(
        self,
        psf,
        dtype=None,
        n_iter=1,
        pre_process=None,
        post_process=None,
        skip_unrolled=False,
        skip_pre=False,
        skip_post=False,
        return_intermediate=False,
        legacy_denoiser=False,
        compensation=None,
        compensation_residual=True,
        psf_network=None,
        psf_residual=True,
        # background subtraction
        direct_background_subtraction=False,
        background_network=None,
        integrated_background_subtraction=False,
        **kwargs,
    ):
        """
        Base constructor. Derived constructor may define new state variables
        here and also reset them in `reset`.

        Parameters
        ----------

        psf : :py:class:`~torch.Tensor`
            Point spread function (PSF) that models forward propagation.
            Must be of shape (depth, height, width, channels) even if
            depth = 1 and channels = 1. You can use :py:func:`~lensless.io.load_psf`
            to load a PSF from a file such that it is in the correct format.
        dtype : float32 or float64
            Data type to use for optimization.
        n_iter : int
            Number of iterations for unrolled algorithm.
        pre_process : :py:class:`function` or :py:class:`~torch.nn.Module`, optional
            If :py:class:`function` : Function to apply to the image estimate before algorithm. Its input most be (image to process, noise_level), where noise_level is a learnable parameter. If it include aditional learnable parameters, they will not be added to the parameter list of the algorithm. To allow for traning, the function must be autograd compatible.
            If :py:class:`~torch.nn.Module` : A DruNet compatible network to apply to the image estimate before algorithm. See ``utils.image.apply_denoiser`` for more details. The network will be included as a submodule of the algorithm and its parameters will be added to the parameter list of the algorithm. If this isn't intended behavior, set requires_grad=False.
        post_process : :py:class:`function` or :py:class:`~torch.nn.Module`, optional
            If :py:class:`function` : Function to apply to the image estimate after the whole algorithm. Its input most be (image to process, noise_level), where noise_level is a learnable parameter. If it include aditional learnable parameters, they will not be added to the parameter list of the algorithm. To allow for traning, the function must be autograd compatible.
            If :py:class:`~torch.nn.Module` : A DruNet compatible network to apply to the image estimate after the whole algorithm. See ``utils.image.apply_denoiser`` for more details. The network will be included as a submodule of the algorithm and its parameters will be added to the parameter list of the algorithm. If this isn't intended behavior, set requires_grad=False.
        skip_unrolled : bool, optional
            Whether to skip the unrolled algorithm and only apply the pre- or post-processor block (e.g. to just use a U-Net for reconstruction).
        return_unrolled_output : bool, optional
            Whether to return the output of the unrolled algorithm if also using a post-processor block.
        compensation : list, optional
            Number of channels for each intermediate output in compensation layer, as in "Robust Reconstruction With Deep Learning to Handle Model Mismatch in Lensless Imaging" (2021).
            Post-processor must be defined if compensation provided.
        compensation_residual : bool, optional
            Whether to use residual connection in compensation layer.
        psf_network : :py:class:`function` or :py:class:`~torch.nn.Module`, optional
            Function or model to apply to PSF prior to camera inversion.
        psf_residual : bool, optional
            Whether to use residual connection in PSF network.
        """

        assert isinstance(psf, torch.Tensor), "PSF must be a torch.Tensor"
        super(TrainableReconstructionAlgorithm, self).__init__(
            psf, dtype=dtype, n_iter=n_iter, **kwargs
        )

        self._legacy_denoiser = legacy_denoiser
        self.input_background = False
        if pre_process is not None:
            self.set_pre_process(pre_process)
        else:
            self.pre_process = None
            self.pre_process_model = None
            self.pre_process_param = None
        if post_process is not None:
            self.set_post_process(post_process)
        else:
            self.post_process = None
            self.post_process_model = None
            self.post_process_param = None
        self.skip_unrolled = skip_unrolled
        self.skip_pre = skip_pre
        self.skip_post = skip_post

        # background subtraction
        self.direct_background_subtraction = direct_background_subtraction
        self.integrated_background_subtraction = integrated_background_subtraction
        self.learned_background_subtraction = False
        self.background_network = None
        if background_network is not None:
            if integrated_background_subtraction:
                assert (
                    direct_background_subtraction is False
                ), "Cannot use direct_background_subtraction and background_network at the same time."
                assert (
                    pre_process is None
                ), "Cannot use pre_process with integrated_background_subtraction."
                self.pre_process = background_network
            else:
                self.learned_background_subtraction = True
                assert (
                    direct_background_subtraction is False
                ), "Cannot use direct_background_subtraction and background_network at the same time."
                self.set_background_network(background_network)

        # PSF network
        self.psf_network = None
        self.psf_residual = psf_residual
        if psf_network is not None:
            self.set_psf_network(psf_network)

        # compensation branch
        self.return_intermediate = return_intermediate
        self.compensation_branch = compensation
        if compensation is not None:
            from lensless.recon.utils import CompensationBranch

            assert (
                post_process is not None
            ), "If compensation_branch is True, post_process must be defined."
            assert (
                len(compensation) == n_iter
            ), "compensation_nc must have the same length as n_iter"
            self.compensation_branch = CompensationBranch(
                compensation, residual=compensation_residual
            )
            self.compensation_branch = self.compensation_branch.to(self._psf.device)

        if self.return_intermediate:
            assert (
                post_process is not None or pre_process is not None
            ), "If return_intermediate is True, post_process or pre_process must be defined."
        if self.skip_unrolled:
            assert (
                post_process is not None or pre_process is not None
            ), "If skip_unrolled is True, pre_process or post_process must be defined."

    def _prepare_process_block(self, process):
        """
        Method for preparing the pre or post process block.
        Parameters
        ----------
            process : :py:class:`function` or :py:class:`~torch.nn.Module`, optional
                Pre or post process block to prepare.
        """
        if isinstance(process, torch.nn.Module):
            # If the post_process is a torch module, we assume it is a DruNet like network.
            from lensless.recon.utils import get_drunet_function, get_drunet_function_v2
            from lensless.recon.restormer import get_restormer_function

            process_model = process
            # handle DataParallel
            if isinstance(process, torch.nn.DataParallel):
                process_type = process.module.__class__.__name__
            else:
                process_type = process.__class__.__name__
            if process_type == "Restormer":
                process_function = get_restormer_function(process_model)
            elif self._legacy_denoiser:
                process_function = get_drunet_function(process_model, mode="train")
            else:
                process_function = get_drunet_function_v2(process_model, mode="train")
        elif process is not None:
            # Otherwise, we assume it is a function.
            assert callable(process), "pre_process must be a callable function"
            process_function = process
            process_model = None
        else:
            process_function = None
            process_model = None

        if process_function is not None:
            process_param = torch.nn.Parameter(torch.tensor([1.0], device=self._psf.device))
        else:
            process_param = None

        return process_function, process_model, process_param

    def set_pre_process(self, pre_process):
        if isinstance(pre_process, torch.nn.DataParallel):
            if hasattr(pre_process.module, "input_background"):
                self.input_background = pre_process.module.input_background
        else:
            if hasattr(pre_process, "input_background"):
                self.input_background = pre_process.input_background
        (
            self.pre_process,
            self.pre_process_model,
            self.pre_process_param,
        ) = self._prepare_process_block(pre_process)

    def set_post_process(self, post_process):
        (
            self.post_process,
            self.post_process_model,
            self.post_process_param,
        ) = self._prepare_process_block(post_process)

    def set_background_network(self, background_network):
        (
            self.background_network,
            self.background_network_model,
            self.background_network_param,
        ) = self._prepare_process_block(background_network)

    def set_psf_network(self, psf_network):
        (
            self.psf_network,
            self.psf_network_model,
            self.psf_network_param,
        ) = self._prepare_process_block(psf_network)

    def freeze_pre_process(self):
        """
        Method for freezing the pre process block.
        """
        if self.pre_process_param is not None:
            self.pre_process_param.requires_grad = False
        if self.pre_process_model is not None:
            for param in self.pre_process_model.parameters():
                param.requires_grad = False

    def freeze_post_process(self):
        """
        Method for freezing the post process block.
        """
        if self.post_process_param is not None:
            self.post_process_param.requires_grad = False
        if self.post_process_model is not None:
            for param in self.post_process_model.parameters():
                param.requires_grad = False

    def unfreeze_pre_process(self):
        """
        Method for unfreezing the pre process block.
        """
        if self.pre_process_param is not None:
            self.pre_process_param.requires_grad = True
        if self.pre_process_model is not None:
            for param in self.pre_process_model.parameters():
                param.requires_grad = True

    def unfreeze_post_process(self):
        """
        Method for unfreezing the post process block.
        """
        if self.post_process_param is not None:
            self.post_process_param.requires_grad = True
        if self.post_process_model is not None:
            for param in self.post_process_model.parameters():
                param.requires_grad = True

    def forward(self, batch, psfs=None, background=None):
        """
        Method for performing iterative reconstruction on a batch of images.
        This implementation is a properly vectorized implementation of FISTA.

        Parameters
        ----------
        batch : :py:class:`~torch.Tensor` of shape (batch, depth, channels, height, width)
            The lensless images to reconstruct.
        psfs : :py:class:`~torch.Tensor` of shape (batch, depth, channels, height, width)
            The lensless images to reconstruct.

        Returns
        -------
        :py:class:`~torch.Tensor` of shape (batch, depth, channels, height, width)
            The reconstructed images.
        """
        self._data = batch
        assert len(self._data.shape) == 5, "batch must be of shape (N, D, C, H, W)"
        batch_size = batch.shape[0]

        if self.direct_background_subtraction:
            assert (
                background is not None
            ), "If direct_background_subtraction is True, background must be defined."
            self._data = self._data - background
            self._data = torch.clamp(self._data, 0, 1)
        elif self.learned_background_subtraction:
            assert (
                background is not None
            ), "If learned_background_subtraction is True, background must be defined."
            assert (
                self.background_network is not None
            ), "If learned_background_subtraction is True, background_network must be defined."

            self._data = self._data - self.background_network(
                background, self.background_network_param
            ).to(self._data.device)
            self._data = torch.clamp(self._data, 0, 1)

        # set / transform PSFs if need be
        if self.psf_network is not None:
            if psfs is None:
                psfs = self._psf.to(self._data.device)
            if self.psf_residual:
                psfs = self.psf_network(psfs, self.psf_network_param).to(psfs.device) + psfs
            else:
                psfs = self.psf_network(psfs, self.psf_network_param).to(psfs.device)

        if psfs is not None:
            # assert same shape
            assert psfs.shape[-3:] == batch.shape[-3:], "psfs must have the same shape as batch"
            # -- update convolver
            self._convolver = RealFFTConvolve2D(psfs.to(self._data.device), **self._convolver_param)
        elif self._data.device != self._convolver._H.device:
            # needed for multi-GPU... TODO better solution?
            self._convolver = RealFFTConvolve2D(
                self._psf.to(self._data.device), **self._convolver_param
            )

        # pre process data
        if self.integrated_background_subtraction:
            # use preprocess for background subtraction
            self._data = self.pre_process(self._data, noise_level=background)
        elif self.pre_process is not None and not self.skip_pre:
            # preproc that doesn't do background subtraction
            device_before = self._data.device
            self._data = self.pre_process(
                self._data,
                noise_level=self.pre_process_param,
                background=background if self.input_background else None,
            )
            self._data = self._data.to(device_before)
        pre_processed = self._data

        self.reset(batch_size=batch_size)

        # unrolled algorithm
        if not self.skip_unrolled:
            if self.compensation_branch is not None:
                compensation_branch_inputs = [self._data]

            for i in range(self._n_iter):
                self._update(i)
                if self.compensation_branch is not None and i < self._n_iter - 1:
                    compensation_branch_inputs.append(self._form_image())

            image_est = self._form_image()
        else:
            image_est = self._data

        # post process data
        if self.post_process is not None and not self.skip_post:
            compensation_output = None
            if self.compensation_branch is not None:
                compensation_output = self.compensation_branch(compensation_branch_inputs)

            final_est = self.post_process(
                image_est,
                noise_level=self.post_process_param,
                compensation_output=compensation_output,
            )
        else:
            final_est = image_est

        if self.return_intermediate:
            return final_est, image_est, pre_processed, psfs
        else:
            return final_est

    def apply(
        self,
        disp_iter=10,
        plot_pause=0.2,
        plot=True,
        save=False,
        gamma=None,
        ax=None,
        reset=True,
        output_intermediate=False,
        background=None,
    ):
        """
        Method for performing iterative reconstruction. Contrary to non-trainable reconstruction
        algorithm, the number of iteration isn't required. Note that `set_data` must be called
        beforehand.

        Parameters
        ----------
        disp_iter : int
            How often to display and/or intermediate reconstruction (in number
            of iterations). If `None` OR `plot` or `save` are False, no
            intermediate reconstruction will be plotted/saved.
        plot_pause : float
            Number of seconds to pause after displaying reconstruction.
        plot : bool
            Whether to plot final result, and intermediate results if
            `disp_iter` is not None.
        save : bool
            Whether to save final result (as PNG), and intermediate results if
            `disp_iter` is not None.
        gamma : float, optional
            Gamma correction factor to apply for plots. Default is None.
        ax : :py:class:`~matplotlib.axes.Axes`, optional
            `Axes` object to fill for plotting/saving, default is to create one.
        output_intermediate : bool, optional
            Whether to output intermediate reconstructions after preprocessing and before postprocessing.

        Returns
        -------
        final_im : :py:class:`~torch.Tensor`
            Final reconstruction.
        ax : :py:class:`~matplotlib.axes.Axes`
            `Axes` object on which final reconstruction is displayed. Only
            returning if `plot` or `save` is True.

        """

        if self.direct_background_subtraction:
            assert (
                background is not None
            ), "If direct_background_subtraction is True, background must be defined."
            self._data = self._data - background
            self._data = torch.clamp(self._data, 0, 1)
        elif self.learned_background_subtraction:
            assert (
                background is not None
            ), "If learned_background_subtraction is True, background must be defined."
            assert (
                self.background_network is not None
            ), "If learned_background_subtraction is True, background_network must be defined."

            self._data = self._data - self.background_network(
                background, self.background_network_param
            ).to(self._data.device)
            self._data = torch.clamp(self._data, 0, 1)

        # transform PSF if need be
        psf = None
        if self.psf_network is not None:
            psf = self._psf.to(self._data.device)
            if self.psf_residual:
                psf = self.psf_network(psf, self.psf_network_param).to(psf.device) + psf
            else:
                psf = self.psf_network(psf, self.psf_network_param).to(psf.device)
        if psf is not None:  # -- update convolver
            self._convolver = RealFFTConvolve2D(psf.to(self._data.device), **self._convolver_param)

        pre_processed_image = None
        if self.integrated_background_subtraction:
            # use preprocess for background subtraction
            self._data = self.pre_process(self._data, noise_level=background)
        elif self.pre_process is not None and not self.skip_pre:

            # preproc that doesn't do background subtraction
            self._data = self.pre_process(
                self._data,
                noise_level=self.pre_process_param,
                background=background if self.input_background else None,
            )
        if self.pre_process is not None and output_intermediate:
            pre_processed_image = self._data[0, ...].clone()

        if not self.skip_unrolled:
            im = super(TrainableReconstructionAlgorithm, self).apply(
                n_iter=self._n_iter,
                disp_iter=disp_iter,
                plot_pause=plot_pause,
                plot=plot,
                save=save,
                gamma=gamma,
                ax=ax,
                reset=reset,
            )
        else:
            im = self._data

        # remove plot if returned
        if plot:
            im, _ = im

        # post process data
        pre_post_process_image = None
        if self.post_process is not None and not self.skip_post:

            compensation_output = None
            if self.compensation_branch is not None:
                compensation_output = self.compensation_branch(self.compensation_branch_inputs)

            # apply post process
            if output_intermediate:
                pre_post_process_image = im.clone()
            im = self.post_process(
                im, noise_level=self.post_process_param, compensation_output=compensation_output
            )

        if plot:
            ax = plot_image(self._get_numpy_data(im[0]), ax=ax, gamma=gamma)
            ax.set_title(
                "Final reconstruction after {} iterations and post process".format(self._n_iter)
            )
            if save:
                plt.savefig(plib.Path(save) / "final.png")

        if output_intermediate:
            return_items = im, pre_post_process_image, pre_processed_image
            if self.psf_network is not None:
                return_items += (psf,)
            return return_items
        elif plot:
            return im, ax
        else:
            return im



# #############################################################################
# gradient_descent.py
# =================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# Julien SAHLI [julien.sahli@epfl.ch]
# #############################################################################


import numpy as np
from lensless.recon.recon import ReconstructionAlgorithm
import inspect
from lensless.utils.io import load_data
import time

try:
    import torch

    torch_available = True
except ImportError:
    torch_available = False


class GradientDescentUpdate:
    """Gradient descent update techniques."""

    VANILLA = "vanilla"
    NESTEROV = "nesterov"
    FISTA = "fista"

    @staticmethod
    def all_values():
        vals = []
        for i in inspect.getmembers(GradientDescentUpdate):
            # remove private and protected functions, and this function
            if not i[0].startswith("_") and not callable(i[1]):
                vals.append(i[1])
        return vals


def non_neg(xi):
    """
    Clip input so that it is non-negative.

    Parameters
    ----------
    xi : :py:class:`~numpy.ndarray`
        Data to clip.

    Returns
    -------
    nonneg : :py:class:`~numpy.ndarray`
        Non-negative projection of input.

    """
    if torch_available and isinstance(xi, torch.Tensor):
        return torch.maximum(xi, torch.zeros_like(xi))
    else:
        return np.maximum(xi, 0)


class GradientDescent(ReconstructionAlgorithm):
    """
    Object for applying projected gradient descent.
    """

    def __init__(self, psf, dtype=None, proj=non_neg, lip_fact=1.8, **kwargs):
        """

        Parameters
        ----------
        psf : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
            Point spread function (PSF) that models forward propagation.
            Must be of shape (depth, height, width, channels) even if
            depth = 1 and channels = 1. You can use :py:func:`~lensless.io.load_psf`
            to load a PSF from a file such that it is in the correct format.
        dtype : float32 or float64
            Data type to use for optimization. Default is float32.
        proj : :py:class:`function`
            Projection function to apply at each iteration. Default is
            non-negative.
        """

        assert callable(proj)
        self._proj = proj
        self._lip_fact = lip_fact
        super(GradientDescent, self).__init__(psf, dtype, **kwargs)

        if self._denoiser is not None:
            print("Using denoiser in gradient descent.")
            # redefine projection function
            self._proj = self._denoiser

    def reset(self):
        if self.is_torch:
            if self._initial_est is not None:
                self._image_est = self._initial_est
            else:
                # initial guess, half intensity image
                psf_flat = self._psf.reshape(-1, self._psf_shape[3])
                pixel_start = (
                    torch.max(psf_flat, axis=0).values + torch.min(psf_flat, axis=0).values
                ) / 2
                # initialize image estimate as [Batch, Depth, Height, Width, Channels]
                self._image_est = torch.ones_like(self._psf[None, ...]) * pixel_start

            # set step size as < 2 / lipschitz
            Hadj_flat = self._convolver._Hadj.reshape(-1, self._psf_shape[3])
            H_flat = self._convolver._H.reshape(-1, self._psf_shape[3])
            self._alpha = torch.real(
                self._lip_fact / torch.max(torch.abs(Hadj_flat * H_flat), axis=0).values
            )

        else:
            if self._initial_est is not None:
                self._image_est = self._initial_est
            else:
                psf_flat = self._psf.reshape(-1, self._psf_shape[3])
                pixel_start = (np.max(psf_flat, axis=0) + np.min(psf_flat, axis=0)) / 2
                # initialize image estimate as [Batch, Depth, Height, Width, Channels]
                self._image_est = np.ones_like(self._psf[None, ...]) * pixel_start

            # set step size as < 2 / lipschitz
            Hadj_flat = self._convolver._Hadj.reshape(-1, self._psf_shape[3])
            H_flat = self._convolver._H.reshape(-1, self._psf_shape[3])
            self._alpha = np.real(self._lip_fact / np.max(Hadj_flat * H_flat, axis=0))

    def _grad(self):
        diff = self._convolver.convolve(self._image_est) - self._data
        return self._convolver.deconvolve(diff)

    def _update(self, iter):
        self._image_est -= self._alpha * self._grad()
        self._image_est = self._form_image()

    def _form_image(self):
        if self._denoiser is not None:
            return self._proj(self._image_est, self._denoiser_noise_level)
        else:
            return self._proj(self._image_est)


class NesterovGradientDescent(GradientDescent):
    """
    Object for applying projected gradient descent with Nesterov momentum for
    acceleration.

    A nice tutorial/ blog post on Nesterov momentum can be found
    `here <https://machinelearningmastery.com/gradient-descent-with-nesterov-momentum-from-scratch/>`_.

    """

    def __init__(self, psf, dtype=None, proj=non_neg, p=0, mu=0.9, **kwargs):
        """

        Parameters
        ----------
        psf : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
            Point spread function (PSF) that models forward propagation.
            Must be of shape (depth, height, width, channels) even if
            depth = 1 and channels = 1. You can use :py:func:`~lensless.io.load_psf`
            to load a PSF from a file such that it is in the correct format.
        dtype : float32 or float64
            Data type to use for optimization. Default is float32.
        proj : :py:class:`function`
            Projection function to apply at each iteration. Default is
            non-negative.
        p : float
            Momentum parameter that keeps track of changes. By default, this
            is initialized to 0.
        mu : float
            Momentum parameter. Default is 0.9.
        """
        self._p = p
        self._mu = mu
        super(NesterovGradientDescent, self).__init__(psf, dtype, proj, **kwargs)

    def reset(self, p=0, mu=0.9):
        self._p = p
        self._mu = mu
        super(NesterovGradientDescent, self).reset()

    def _update(self, iter):
        p_prev = self._p
        self._p = self._mu * self._p - self._alpha * self._grad()
        self._image_est += -self._mu * p_prev + (1 + self._mu) * self._p
        # self._image_est = self._proj(self._image_est)
        self._image_est = self._form_image()


class FISTA(GradientDescent):
    """
    Object for applying projected gradient descent with FISTA (Fast Iterative
    Shrinkage-Thresholding Algorithm) for acceleration.

    Paper: https://www.ceremade.dauphine.fr/~carlier/FISTA

    """

    def __init__(self, psf, dtype=None, proj=non_neg, tk=1.0, **kwargs):
        """

        Parameters
        ----------
        psf : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
            Point spread function (PSF) that models forward propagation.
            Must be of shape (depth, height, width, channels) even if
            depth = 1 and channels = 1. You can use :py:func:`~lensless.io.load_psf`
            to load a PSF from a file such that it is in the correct format.
        dtype : float32 or float64
            Data type to use for optimization. Default is float32.
        proj : :py:class:`function`
            Projection function to apply at each iteration. Default is
            non-negative.
        tk : float
            Initial step size parameter for FISTA. It is updated at each iteration
            according to Eq. 4.2 of paper. By default, initialized to 1.0.

        """
        self._initial_tk = tk

        super(FISTA, self).__init__(psf, dtype, proj, **kwargs)

        self._tk = tk
        self._xk = self._image_est

    def reset(self, tk=None):
        super(FISTA, self).reset()
        if tk:
            self._tk = tk
        else:
            self._tk = self._initial_tk
        self._xk = self._image_est

    def _update(self, iter):
        self._image_est -= self._alpha * self._grad()
        xk = self._form_image()
        tk = (1 + np.sqrt(1 + 4 * self._tk**2)) / 2
        self._image_est = xk + (self._tk - 1) / tk * (xk - self._xk)
        self._tk = tk
        self._xk = xk


def apply_gradient_descent(psf_fp, data_fp, n_iter, verbose=False, proj=non_neg, **kwargs):

    # load data
    psf, data = load_data(psf_fp=psf_fp, data_fp=data_fp, plot=False, **kwargs)

    # create reconstruction object
    recon = GradientDescent(psf, n_iter=n_iter, proj=proj)

    # set data
    recon.set_data(data)

    # perform reconstruction
    start_time = time.time()
    res = recon.apply(plot=False)
    proc_time = time.time() - start_time

    if verbose:
        print(f"Reconstruction time : {proc_time} s")
        print(f"Reconstruction shape: {res.shape}")
    return res



# #############################################################################
# io.py
# =================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################


import os.path
import warnings

import cv2
import numpy as np
from PIL import Image

from lensless.hardware.constants import RPI_HQ_CAMERA_BLACK_LEVEL, RPI_HQ_CAMERA_CCM_MATRIX
from lensless.utils.image import bayer2rgb_cc, print_image_info, resize, rgb2gray, get_max_val
from lensless.utils.plot import plot_image


def load_image(
    fp,
    verbose=False,
    flip=False,
    flip_ud=False,
    flip_lr=False,
    bayer=False,
    black_level=RPI_HQ_CAMERA_BLACK_LEVEL,
    blue_gain=None,
    red_gain=None,
    ccm=RPI_HQ_CAMERA_CCM_MATRIX,
    back=None,
    nbits_out=None,
    as_4d=False,
    downsample=None,
    bg=None,
    return_float=False,
    shape=None,
    dtype=None,
    normalize=True,
    bgr_input=True,
):
    """
    Load image as numpy array.

    Parameters
    ----------
    fp : str
        Full path to file.
    verbose : bool, optional
        Whether to plot into about file.
    flip : bool
        Whether to flip data (vertical and horizontal).
    bayer : bool
        Whether input data is Bayer.
    blue_gain : float
        Blue gain for color correction.
    red_gain : float
        Red gain for color correction.
    black_level : float
        Black level. Default is to use that of Raspberry Pi HQ camera.
    ccm : :py:class:`~numpy.ndarray`
        Color correction matrix. Default is to use that of Raspberry Pi HQ camera.
    back : array_like
        Background level to subtract.
    nbits_out : int
        Output bit depth. Default is to use that of input.
    as_4d : bool
        Add depth and color dimensions if necessary so that image is 4D: (depth,
        height, width, color).
    downsample : int, optional
        Downsampling factor. Recommended for image reconstruction.
    bg : array_like
        Background level to subtract.
    return_float : bool
        Whether to return image as float array, or unsigned int.
    shape : tuple, optional
        Shape (H, W, C) to resize to.
    dtype : str, optional
        Data type of returned data. Default is to use that of input.
    normalize : bool, default True
        If ``return_float``, whether to normalize data to maximum value of 1.

    Returns
    -------
    img : :py:class:`~numpy.ndarray`
        RGB image of dimension (height, width, 3).
    """
    assert os.path.isfile(fp)

    nbits = None  # input bit depth
    if "dng" in fp:
        import rawpy

        assert bayer
        raw = rawpy.imread(fp)
        img = raw.raw_image
        # # # TODO : use raw.postprocess? to much unknown processing...
        # img = raw.postprocess(
        #     adjust_maximum_thr=0,  # default 0.75
        #     no_auto_scale=False,
        #     # no_auto_scale=True,
        #     gamma=(1, 1),
        #     bright=1,  # default 1
        #     exp_shift=1,
        #     no_auto_bright=True,
        #     # use_camera_wb=True,
        #     # use_auto_wb=False,
        #     # -- gives better balance for PSF measurement
        #     use_camera_wb=False,
        #     use_auto_wb=True,  # default is False? f both use_camera_wb and use_auto_wb are True, then use_auto_wb has priority.
        # )

        # if red_gain is None or blue_gain is None:
        #     camera_wb = raw.camera_whitebalance
        #     red_gain = camera_wb[0]
        #     blue_gain = camera_wb[1]

        nbits = int(np.ceil(np.log2(raw.white_level)))
        ccm = raw.color_matrix[:, :3]
        black_level = np.array(raw.black_level_per_channel[:3]).astype(np.float32)
    elif "npy" in fp or "npz" in fp:
        img = np.load(fp)
    else:
        img = cv2.imread(fp, cv2.IMREAD_UNCHANGED)

    if bayer:
        assert len(img.shape) == 2, img.shape
        if nbits is None:
            if img.max() > 255:
                # HQ camera
                nbits = 12
            else:
                nbits = 8

        if back:
            back_img = cv2.imread(back, cv2.IMREAD_UNCHANGED)
            dtype = img.dtype
            img = img.astype(np.float32) - back_img.astype(np.float32)
            img = np.clip(img, a_min=0, a_max=img.max())
            img = img.astype(dtype)
        if nbits_out is None:
            nbits_out = nbits

        img = bayer2rgb_cc(
            img,
            nbits=nbits,
            blue_gain=blue_gain,
            red_gain=red_gain,
            black_level=black_level,
            ccm=ccm,
            nbits_out=nbits_out,
        )

    else:
        if len(img.shape) == 3 and bgr_input:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

    original_dtype = img.dtype

    if flip:
        img = np.flipud(img)
        img = np.fliplr(img)
    if flip_ud:
        img = np.flipud(img)
    if flip_lr:
        img = np.fliplr(img)

    if bg is not None:

        # if bg is float vector, turn into int-valued vector
        if bg.max() <= 1 and img.dtype not in [np.float32, np.float64]:
            bg = bg * get_max_val(img)

        img = img - bg
        img = np.clip(img, a_min=0, a_max=img.max())

    if as_4d:
        if len(img.shape) == 3:
            img = img[np.newaxis, :, :, :]
        elif len(img.shape) == 2:
            img = img[np.newaxis, :, :, np.newaxis]

    if downsample is not None or shape is not None:
        if downsample is not None:
            factor = 1 / downsample
        else:
            factor = None
        img = resize(img, factor=factor, shape=shape)

    if return_float:
        if dtype is None:
            dtype = np.float32
        assert dtype == np.float32 or dtype == np.float64
        img = img.astype(dtype)
        if normalize:
            img /= img.max()

    else:
        if dtype is None:
            dtype = original_dtype
        img = img.astype(dtype)

    if verbose:
        print_image_info(img)

    return img


def load_psf(
    fp,
    downsample=1,
    return_float=True,
    bg_pix=(5, 25),
    return_bg=False,
    flip=False,
    flip_ud=False,
    flip_lr=False,
    verbose=False,
    bayer=False,
    blue_gain=None,
    red_gain=None,
    dtype=np.float32,
    nbits_out=None,
    single_psf=False,
    shape=None,
    use_3d=False,
    bgr_input=True,
    force_rgb=False,
):
    """
    Load and process PSF for analysis or for reconstruction.

    Basic steps are:
    * Load image.
    * (Optionally) subtract background. Recommended.
    * (Optionally) resize to more manageable size
    * (Optionally) normalize within [0, 1] if using for reconstruction; otherwise cast back to uint for analysis.

    Parameters
    ----------
    fp : str
        Full path to file.
    downsample : int, optional
        Downsampling factor. Recommended for image reconstruction.
    return_float : bool, optional
        Whether to return PSF as float array, or unsigned int.
    bg_pix : tuple, optional
        Section of pixels to take from top left corner to remove background level. Set to `None` to omit this
        step, althrough it is highly recommended.
    return_bg : bool, optional
        Whether to return background level, for removing from data for reconstruction.
    flip : bool, optional
        Whether to flip up-down and left-right.
    verbose : bool
        Whether to print metadata.
    bayer : bool
        Whether input data is Bayer.
    blue_gain : float
        Blue gain for color correction.
    red_gain : float
        Red gain for color correction.
    dtype : float32 or float64
        Data type of returned data.
    nbits_out : int
        Output bit depth. Default is to use that of input.
    single_psf : bool
        Whether to sum RGB channels into single PSF, same across channels. Done
        in "Learned reconstructions for practical mask-based lensless imaging"
        of Kristina Monakhova et. al.

    Returns
    -------
    psf : :py:class:`~numpy.ndarray`
        4-D array of PSF.
    """

    # load image data and extract necessary channels
    if use_3d:
        assert os.path.isfile(fp)
        if fp.endswith(".npy"):
            psf = np.load(fp)
        elif fp.endswith(".npz"):
            archive = np.load(fp)
            if len(archive.files) > 1:
                print("Warning: more than one array in .npz archive, using first")
            elif len(archive.files) == 0:
                raise ValueError("No arrays in .npz archive")
            psf = np.load(fp)[archive.files[0]]
        else:
            raise ValueError("File format not supported")
    else:
        psf = load_image(
            fp,
            verbose=False,
            flip=flip,
            flip_ud=flip_ud,
            flip_lr=flip_lr,
            bayer=bayer,
            blue_gain=blue_gain,
            red_gain=red_gain,
            nbits_out=nbits_out,
            bgr_input=bgr_input,
        )

    original_dtype = psf.dtype
    max_val = get_max_val(psf)
    psf = np.array(psf, dtype=dtype)

    if force_rgb:
        if len(psf.shape) == 2:
            psf = np.stack([psf] * 3, axis=2)
        elif len(psf.shape) == 3:
            pass

    if use_3d:
        if len(psf.shape) == 3:
            grayscale = True
            psf = psf[:, :, :, np.newaxis]
        else:
            assert len(psf.shape) == 4
            grayscale = False

    else:
        if len(psf.shape) == 3:
            grayscale = False
            psf = psf[np.newaxis, :, :, :]
        else:
            assert len(psf.shape) == 2
            grayscale = True
            psf = psf[np.newaxis, :, :, np.newaxis]

    # check that all depths of the psf have the same shape.
    for i in range(len(psf)):
        assert psf[0].shape == psf[i].shape

    # subtract background, assume black edges
    if bg_pix is None:
        bg = np.zeros(len(np.shape(psf)))

    else:
        # grayscale
        if grayscale:
            bg = np.mean(psf[:, bg_pix[0] : bg_pix[1], bg_pix[0] : bg_pix[1], :])
            psf -= bg

        # rgb
        else:
            bg = []
            for i in range(psf.shape[3]):
                bg_i = np.mean(psf[:, bg_pix[0] : bg_pix[1], bg_pix[0] : bg_pix[1], i])
                psf[:, :, :, i] -= bg_i
                bg.append(bg_i)

        psf = np.clip(psf, a_min=0, a_max=psf.max())
        bg = np.array(bg)

    # resize
    if downsample != 1 or shape is not None:
        psf = resize(psf, shape=shape, factor=1 / downsample)

    if single_psf:
        if not grayscale:
            # TODO : in Lensless Learning, they sum channels --> `psf_diffuser = np.sum(psf_diffuser,2)`
            # https://github.com/Waller-Lab/LenslessLearning/blob/master/pre-trained%20reconstructions.ipynb
            psf = np.sum(psf, axis=3)
            psf = psf[:, :, :, np.newaxis]
        else:
            warnings.warn("Notice : single_psf has no effect for grayscale psf")
            single_psf = False

    # normalize
    if return_float:
        # psf /= psf.max()
        psf /= np.linalg.norm(psf.ravel())
        bg /= max_val
    else:
        psf = psf.astype(original_dtype)

    if verbose:
        print_image_info(psf)

    if return_bg:
        return psf, bg
    else:
        return psf


def load_data(
    psf_fp,
    data_fp,
    background_fp=None,
    return_bg=False,
    remove_background=True,
    return_float=True,
    downsample=None,
    bg_pix=(5, 25),
    plot=True,
    flip=False,
    flip_ud=False,
    flip_lr=False,
    bayer=False,
    blue_gain=None,
    red_gain=None,
    gamma=None,
    gray=False,
    dtype=None,
    single_psf=False,
    shape=None,
    use_torch=False,
    torch_device="cpu",
    normalize=False,
    bgr_input=True,
):
    """
    Load data for image reconstruction.

    Parameters
    ----------
    psf_fp : str
        Full path to PSF file.
    data_fp : str
        Full path to measurement file.
    return_float : bool, optional
        Whether to return PSF as float array, or unsigned int.
    downsample : int or float
        Downsampling factor.
    bg_pix : tuple, optional
        Section of pixels to take from top left corner to remove background
        level. Set to `None` to omit this step, although it is highly
        recommended.
    plot : bool, optional
        Whether or not to plot PSF and raw data.
    flip : bool
        Whether to flip data (vertical and horizontal).
    bayer : bool
        Whether input data is Bayer.
    blue_gain : float
        Blue gain for color correction.
    red_gain : float
        Red gain for color correction.
    gamma : float, optional
        Optional gamma factor to apply, ONLY for plotting. Default is None.
    gray : bool
        Whether to load as grayscale or RGB.
    dtype : float32 or float64, default float32
        Data type of returned data.
    single_psf : bool
        Whether to sum RGB channels into single PSF, same across channels. Done
        in "Learned reconstructions for practical mask-based lensless imaging"
        of Kristina Monakhova et. al.
    normalize : bool default True
        Whether to normalize data to maximum value of 1.

    Returns
    -------
    psf : :py:class:`~numpy.ndarray`
        2-D array of PSF.
    data : :py:class:`~numpy.ndarray`
        2-D array of raw measurement data.
    """

    assert os.path.isfile(psf_fp)
    assert os.path.isfile(data_fp)
    if shape is None:
        assert downsample is not None

    if dtype is None:
        dtype = np.float32
    elif dtype == "float32":
        dtype = np.float32
    elif dtype == "float64":
        dtype = np.float64
    else:
        raise ValueError("dtype must be float32 or float64")

    use_3d = psf_fp.endswith(".npy") or psf_fp.endswith(".npz")

    # load and process PSF data
    bg = None
    res = load_psf(
        psf_fp,
        downsample=downsample,
        return_float=return_float,
        bg_pix=bg_pix,
        return_bg=True if bg_pix is not None else False,
        flip=flip,
        flip_ud=flip_ud,
        flip_lr=flip_lr,
        bayer=bayer,
        blue_gain=blue_gain,
        red_gain=red_gain,
        dtype=dtype,
        single_psf=single_psf,
        shape=shape,
        use_3d=use_3d,
        bgr_input=bgr_input,
    )
    if bg_pix is not None:
        psf, bg = res
    else:
        psf = res

    # load and process raw measurement
    data = load_image(
        data_fp,
        flip=flip,
        flip_ud=flip_ud,
        flip_lr=flip_lr,
        bayer=bayer,
        blue_gain=blue_gain,
        red_gain=red_gain,
        bg=bg,
        as_4d=True,
        return_float=return_float,
        shape=shape,
        normalize=normalize if background_fp is None else False,
        bgr_input=bgr_input,
    )

    if background_fp is not None:
        bg = load_image(
            background_fp,
            flip=flip,
            bayer=bayer,
            blue_gain=blue_gain,
            red_gain=red_gain,
            as_4d=True,
            return_float=return_float,
            shape=shape,
            normalize=False,
            bgr_input=bgr_input,
        )
        assert bg.shape == data.shape

        if remove_background:
            data -= bg

        # clip to 0
        data = np.clip(data, a_min=0, a_max=data.max())

        if normalize:
            data /= data.max()
            bg /= data.max()  # to normalize by the same factor

    if data.shape != psf.shape:
        # in DiffuserCam dataset, images are already reshaped
        data = resize(data, shape=psf.shape)

        if background_fp is not None:
            bg = resize(bg, shape=psf.shape)

    if data.shape[3] > 1 and psf.shape[3] == 1:
        warnings.warn(
            "Warning: loaded a grayscale PSF with RGB data. Repeating PSF across channels."
            "This may be an error as the PSF and the data are likely from different datasets."
        )
        psf = np.repeat(psf, data.shape[3], axis=3)

    if data.shape[3] == 1 and psf.shape[3] > 1:
        warnings.warn(
            "Warning: loaded a RGB PSF with grayscale data. Repeating data across channels."
            "This may be an error as the PSF and the data are likely from different datasets."
        )
        data = np.repeat(data, psf.shape[3], axis=3)

    if data.shape[3] != psf.shape[3]:
        raise ValueError(
            "PSF and data must have same number of channels, check that they are from the same dataset."
        )

    if gray:
        psf = np.array(rgb2gray(psf), np.newaxis)
        data = np.array(rgb2gray(data), np.newaxis)

    if plot:
        ax = plot_image(psf[0], gamma=gamma)
        ax.set_title("PSF of the first depth")
        ax = plot_image(data[0], gamma=gamma)
        ax.set_title("Raw data")

    psf = np.array(psf, dtype=dtype)
    data = np.array(data, dtype=dtype)
    bg = np.array(bg, dtype=dtype)
    if use_torch:
        import torch

        if dtype == np.float32:
            torch_dtype = torch.float32
        elif dtype == np.float64:
            torch_dtype = torch.float64

        psf = torch.from_numpy(psf).type(torch_dtype).to(torch_device)
        data = torch.from_numpy(data).type(torch_dtype).to(torch_device)
        bg = torch.from_numpy(bg).type(torch_dtype).to(torch_device)

    if return_bg:
        return psf, data, bg
    else:
        return psf, data


def save_image(img, fp, max_val=255, normalize=True):
    """Save as uint8 image."""

    img_tmp = img.copy()

    if normalize:

        if img_tmp.dtype == np.uint16 or img_tmp.dtype == np.uint8:
            img_tmp = img_tmp.astype(np.float32)

        img_tmp -= img_tmp.min()
        img_tmp /= img_tmp.max()
        img_tmp *= max_val
        img_tmp = img_tmp.astype(np.uint8)

    else:

        if img_tmp.dtype == np.float64 or img_tmp.dtype == np.float32:
            # check within [0, 1] and convert to uint8

            normalized = False
            if img_tmp.min() < 0:
                img_tmp -= img_tmp.min()
                normalized = True
            if img_tmp.max() > 1:
                img_tmp /= img_tmp.max()
                normalized = True
            if normalized:
                print(f"Warning (out of range): {fp} normalizing data to [0, 1]")
            img_tmp *= max_val
            img_tmp = img_tmp.astype(np.uint8)

    # save
    if len(img_tmp.shape) == 3 and img_tmp.shape[2] == 3:
        # RGB
        img_tmp = Image.fromarray(img_tmp)
    else:
        # grayscale
        img_tmp = Image.fromarray(img_tmp.squeeze())
    img_tmp.save(fp)


def get_dtype(dtype=None, is_torch=False):
    """
    Get dtype for numpy or torch.

    Parameters
    ----------
    dtype : str, optional
        "float32" or "float64", Default is "float32".
    is_torch : bool, optional
        Whether to return torch dtype.
    """
    if dtype is None:
        dtype = "float32"
    assert dtype == "float32" or dtype == "float64"

    if is_torch:
        import torch

    if dtype is None:
        if is_torch:
            dtype = torch.float32
        else:
            dtype = np.float32
    else:
        if is_torch:
            dtype = torch.float32 if dtype == "float32" else torch.float64
        else:
            dtype = np.float32 if dtype == "float32" else np.float64

    return dtype


def get_ctypes(dtype, is_torch):
    if not is_torch:
        if dtype == np.float32 or dtype == np.complex64:
            return np.complex64, np.complex64
        elif dtype == np.float64 or dtype == np.complex128:
            return np.complex128, np.complex128
        else:
            raise ValueError("Unexpected dtype: ", dtype)
    else:
        import torch

        if dtype == np.float32 or dtype == np.complex64:
            return torch.complex64, np.complex64
        elif dtype == np.float64 or dtype == np.complex128:
            return torch.complex128, np.complex128
        elif dtype == torch.float32 or dtype == torch.complex64:
            return torch.complex64, np.complex64
        elif dtype == torch.float64 or dtype == torch.complex128:
            return torch.complex128, np.complex128
        else:
            raise ValueError("Unexpected dtype: ", dtype)



# #############################################################################
# plot.py
# =================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# Julien SAHLI [julien.sahli@epfl.ch]
# #############################################################################


import numpy as np
import warnings
import matplotlib.pyplot as plt
import os
import json

from lensless.utils.image import FLOAT_DTYPES, get_max_val, gamma_correction, autocorr2d


def plot_image(img, ax=None, gamma=None, normalize=True):
    """
    Plot image data.

    Parameters
    ----------
    img : :py:class:`~numpy.ndarray`
        Data to plot.
    ax : :py:class:`~matplotlib.axes.Axes`, optional
        `Axes` object to fill for plotting/saving, default is to create one.
    gamma : float, optional
        Gamma correction factor to apply for plots. Default is None.
    normalize : bool, optional
        Whether to normalize data to maximum range. Default is True.

    Returns
    -------
    ax : :py:class:`~matplotlib.axes.Axes`
        Axes on which image is plot.
    """

    # if we have only 1 depth, remove the axis
    if img.shape[0] == 1:
        img = img[0]

    # if we have only 1 color channel, remove the axis
    if img.shape[-1] == 1:
        img = img[..., 0]

    disp_img = None
    cmap = None

    # full 3D RGB format : [depth, width, height, color]
    is_3d = False
    if len(img.shape) == 4:
        disp_img = [np.sum(img, axis=axis) for axis in range(3)]
        cmap = None
        is_3d = True

    # data of length 3 means we have to infer whichever depth or color is missing, based on shape.
    elif len(img.shape) == 3:
        if img.shape[2] == 3:  # 2D rgb
            disp_img = [img]
            cmap = None

        else:  # 3D grayscale
            disp_img = [np.sum(img, axis=axis) for axis in range(3)]
            cmap = "gray"
            is_3d = True

    # data of length 2 means we have only width and height
    elif len(img.shape) == 2:  # 2D grayscale
        disp_img = [img]
        cmap = "gray"

    else:
        raise ValueError(f"Unexpected data shape : {img.shape}")

    max_val = [d.max() for d in disp_img]

    if not normalize:
        for i in range(len(max_val)):
            if disp_img[i].dtype not in FLOAT_DTYPES:
                max_val[i] = get_max_val(disp_img[i])
            else:
                max_val[i] = 1

    assert len(disp_img) == 1 or len(disp_img) == 3

    # need float image for gamma correction and plotting
    img_norm = disp_img.copy()
    for i in range(len(img_norm)):
        img_norm[i] = disp_img[i] / max_val[i]
        if gamma and gamma > 1:
            img_norm[i] = gamma_correction(img_norm[i], gamma=gamma)

    if ax is None:
        if not is_3d:
            _, ax = plt.subplots()
        else:
            _, ax = plt.subplots(2, 2)
    else:
        if is_3d:
            assert len(ax) == 2
            assert len(ax[0]) == 2

    if len(img_norm) == 1:
        ax.imshow(img_norm[0], cmap=cmap)

    else:

        # plot each projection separately
        ax[0, 0].imshow(img_norm[0], cmap=cmap)
        ax[0, 1].imshow(np.swapaxes(img_norm[2], 0, 1), cmap=cmap)
        ax[1, 0].imshow(img_norm[1], cmap=cmap)
        ax[1, 1].axis("off")
        ax[0, 1].set_xlabel("Depth")
        ax[1, 0].set_ylabel("Depth")

    return ax


def pixel_histogram(img, nbits=None, ax=None, log_scale=True):
    """
    Plot pixel value histogram.

    Parameters
    ----------
    img : py:class:`~numpy.ndarray`
        2D or 3D image.
    nbits : int, optional
        Bit-depth of camera data.
    ax : :py:class:`~matplotlib.axes.Axes`, optional
            `Axes` object to fill, default is to create one.
    log_scale : bool, optional
        Whether to use log scale in counting number of pixels.

    Return
    ------
    ax : :py:class:`~matplotlib.axes.Axes`
        Axes on which histogram is plot.
    """
    if ax is None:
        _, ax = plt.subplots()

    if nbits:
        # max_val = get_max_val(img, nbits)
        max_val = 2**nbits - 1
    else:
        max_val = int(img.max())

    if len(img.shape) == 3:
        # 3D image
        color_order = ("r", "g", "b")
        for i, col in enumerate(color_order):
            hist, bins = np.histogram(img[:, :, i].ravel(), bins=max_val, range=[0, max_val + 1])
            ax.plot(hist, color=col)
    else:
        # 2D image
        vals = img.flatten()
        hist, bins = np.histogram(vals, bins=max_val, range=[0, max_val + 1])
        ax.plot(hist, color="gray")
    ax.set_xlim([max_val - 1.1 * max_val, max_val * 1.1])

    if log_scale:
        ax.set_yscale("log")
    ax.set_xlabel("Pixel value")
    ax.grid()

    return ax


def plot_cross_section(
    vals,
    idx=None,
    ax=None,
    dB=True,
    plot_db_drop=3,
    min_val=1e-4,
    max_val=None,
    plot_width=None,
    **kwargs,
):
    """
    Plot cross-section of a 2-D image.

    Parameters
    ----------
    vals : py:class:`~numpy.ndarray`
        2-D image data.
    idx : int, optional
        Row for which to plot cross-section. Default is to take middle.
    ax : :py:class:`~matplotlib.axes.Axes`, optional
        `Axes` object to fill, default is to create one.
    dB : bool, optional
        Whether to plot in dB scale.

    Return
    ------
    ax : :py:class:`~matplotlib.axes.Axes`
        Axes on which cross-section is plot.
    """

    if ax is None:
        _, ax = plt.subplots()

    # get cross-section
    if idx is None:
        # if no index, take cross-section with maximum value
        max_idx = np.unravel_index(np.argmax(vals, axis=None), vals.shape)
        idx = max_idx[0]

    cross_section = vals[idx, :].astype(np.float32)

    # normalize
    if max_val is None:
        max_val = cross_section.max()
    cross_section /= max_val
    min_val = max(min_val, cross_section.min())

    if dB:
        cross_section[cross_section < min_val] = min_val
        cross_section = 10 * np.log10(cross_section)
        min_val = 10 * np.log10(min_val)
        ax.set_ylabel("dB")
    x_vals = np.arange(len(cross_section))
    x_vals -= np.argmax(cross_section)
    ax.plot(x_vals, cross_section, **kwargs)
    ax.set_ylim([min_val, 0])
    if plot_width is not None:
        half_width = plot_width // 2 + 1
        ax.set_xlim([-half_width, half_width])
    ax.grid()

    ax.set_title("Cross-section")
    if dB and plot_db_drop:
        cross_section -= np.max(cross_section)
        zero_crossings = np.where(np.diff(np.signbit(cross_section + plot_db_drop)))[0]
        if len(zero_crossings) >= 2:
            zero_crossings -= np.argmax(cross_section)

            # width = zero_crossings[-1] - zero_crossings[0]
            # ax.axvline(x=zero_crossings[0], c="k", linestyle="--")
            # ax.axvline(x=zero_crossings[-1], c="k", linestyle="--")

            first_crossing = np.abs(zero_crossings[np.argmin(np.abs(zero_crossings))])
            width = 2 * np.abs(first_crossing)
            ax.axvline(x=-first_crossing, c="k", linestyle="--")
            ax.axvline(x=+first_crossing, c="k", linestyle="--")
            print(f"-{plot_db_drop}dB width = {width} pixels")
            ax.set_title("Cross-section")
            ax.set_xlabel(f"-{plot_db_drop}dB width = {width}")

        else:
            warnings.warn(
                "Width could not be determined. Did not detect two -{} points : {}".format(
                    plot_db_drop, zero_crossings
                )
            )

    return ax, cross_section


def plot_autocorr2d(vals, pad_mode="reflect", ax=None):
    """
    Plot 2-D autocorrelation of image.

    Parameters
    ----------
    vals : py:class:`~numpy.ndarray`
        2-D image.
    pad_mode : str
        Desired padding. See NumPy documentation: https://numpy.org/doc/stable/reference/generated/numpy.pad.html
    ax : :py:class:`~matplotlib.axes.Axes`, optional
            `Axes` object to fill, default is to create one.

    Return
    ------
    ax : :py:class:`~matplotlib.axes.Axes`
        Axes on which auto-correlation is plot.
    autocorr : py:class:`~numpy.ndarray`
        Auto-correlation.
    """

    nbit_plot = 8
    max_val_plot = 2**nbit_plot - 1

    # compute autocorrelation
    autocorr = autocorr2d(vals, pad_mode=pad_mode)

    # rescale for plotting
    data = autocorr - np.min(autocorr)
    data = data / np.max(np.abs(data))  # normalize the data to 0 - 1
    data = max_val_plot * data  # Now scale by bit depth
    autocorr_img = data.astype(np.uint8)

    if ax is None:
        _, ax = plt.subplots()

    ax.imshow(autocorr_img, cmap="gray", vmin=0, vmax=max_val_plot)
    ax.axis("off")
    return ax, autocorr


def plot_autocorr_rgb(img, width=3, figsize=None, plot_psf=False, psf_gamma=2.2, verbose=False):
    """
    Plot autocorrelation of each channel of an image.

    Parameters
    ----------
    img : py:class:`~numpy.ndarray`
        2-D image.
    width : int, optional
        Width of cross-section to plot. Default is 3dB.

    """

    assert len(img.shape) == 3, "Image must be 3D"
    assert img.shape[2] == 3, "Image must have 3 color channels"

    if plot_psf:
        _, ax_auto = plt.subplots(ncols=3, nrows=3, num="Autocorrelations", figsize=figsize)
    else:
        _, ax_auto = plt.subplots(ncols=3, nrows=2, num="Autocorrelations", figsize=figsize)

    for i, c in enumerate(["Red", "Green", "Blue"]):
        if plot_psf:
            plot_image(
                img[:, :, i],
                ax=ax_auto[0][i],
                gamma=psf_gamma,
                normalize=True,
            )
            # ax_auto[0][i].imshow(img[:, :, i], cmap="gray")
            ax_auto[0][i].axis("off")
            ax_auto[0][i].set_title(f"{c} PSF")

        # plot autocorrelation
        _, autocorr_c = plot_autocorr2d(img[:, :, i], ax=ax_auto[1 if plot_psf else 0][i])
        ax_auto[1][i].set_title("Autocorrelation")
        # # -- horizontal cross-section
        max_idx = np.unravel_index(np.argmax(autocorr_c, axis=None), autocorr_c.shape)
        idx = max_idx[0]
        # ax_auto[1][i].axhline(y=idx, c=c, linestyle="--")

        ax, cross_section = plot_cross_section(
            autocorr_c,
            idx=idx,
            color=c,
            ax=ax_auto[2 if plot_psf else 1][i],
            plot_db_drop=width,
        )
        if verbose:
            print(f"Maximum drop in {c} channel: {cross_section.max() - cross_section.min()}")
        if i != 0:
            ax.set_ylabel("")
    return ax


def compare_models(model_paths, max_epoch=None, linewidth=2, fontsize=18, metrics=None):
    """
    Plot train and test loss for multiple models, and print metrics for best epoch.

    Parameters
    ----------
    model_paths : dict
        Dictionary of model names and their paths.
    max_epoch : int, optional
        Maximum epoch to plot. Default is None.
    linewidth : int, optional
        Line width for plot. Default is 2.
    fontsize : int, optional
        Font size for plot. Default is 18.
    metrics : list, optional
        List of metrics to print. Default is ["PSNR", "SSIM", "LPIPS_Vgg"].
    """

    if metrics is None:
        metrics = ["PSNR", "SSIM", "LPIPS_Vgg"]

    # plot train and test loss
    import matplotlib.colors as mcolors

    plot_colors = list(mcolors.TABLEAU_COLORS.keys())

    _, ax = plt.subplots()
    for model in model_paths:
        model_path = model_paths[model]
        _metrics_path = os.path.join(model_path, "metrics.json")

        assert os.path.exists(_metrics_path), f"Path {_metrics_path} does not exist"
        _test_metrics = json.load(open(_metrics_path))

        color = plot_colors.pop()
        train_loss = np.array(_test_metrics["LOSS"])
        if max_epoch is not None:
            train_loss = train_loss[: max_epoch + 1]
        ax.plot(
            train_loss, label=model + " (train)", color=color, linestyle="--", linewidth=linewidth
        )

        test_loss = np.array(_test_metrics["MSE"]) + np.array(_test_metrics["LPIPS_Vgg"])
        if max_epoch is not None:
            test_loss = test_loss[: max_epoch + 1]
        ax.plot(test_loss, label=model + " (test)", linestyle="-", color=color, linewidth=linewidth)

        # best_epoch = np.argmin(test_loss)
        best_epoch = _test_metrics["best_epoch"]
        print(f"\n-- {model} --")
        print(f"Best epoch for {model}: {best_epoch} / {len(test_loss)-1}")
        print(f"Best test loss for {model}: {test_loss[best_epoch]}")
        # print metrics
        for _metric in metrics:
            print(f"{_metric}: {np.array(_test_metrics[_metric])[best_epoch]:.3}")

    # set font size
    ax.tick_params(axis="both", which="major", labelsize=fontsize)
    ax.set_xlabel("Epoch", fontsize=fontsize)
    ax.set_title("Train-test loss", fontsize=fontsize)

    # legend outside
    ax.legend(loc="upper right", fontsize=fontsize)
    # ax.set_ylim([0.4, 1]);
    if max_epoch is not None:
        ax.set_xlim([0, max_epoch])

    ax.grid()
    return ax



# #############################################################################
# mask.py
# =================
# Authors :
# Aaron FARGEON [aa.fargeon@gmail.com]
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################

"""
Mask Design
===========

This module provides utilities to create different types of masks (:py:class:`~lensless.hardware.mask.CodedAperture`,
:py:class:`~lensless.hardware.mask.PhaseContour`,
:py:class:`~lensless.hardware.mask.FresnelZoneAperture`) and simulate the corresponding PSF.

"""


import abc
import warnings
import numpy as np
import cv2 as cv
from math import sqrt
from perlin_numpy import generate_perlin_noise_2d
from sympy.ntheory import quadratic_residues
from scipy.signal import max_len_seq
from scipy.linalg import circulant
from numpy.linalg import multi_dot
from waveprop.fresnel import fresnel_conv
from waveprop.rs import angular_spectrum
from waveprop.noise import add_shot_noise
from lensless.hardware.sensor import VirtualSensor
from lensless.utils.image import resize
import matplotlib.pyplot as plt

try:
    import torch

    torch_available = True
except ImportError:
    torch_available = False


class Mask(abc.ABC):
    """
    Parent ``Mask`` class. Attributes common to each type of mask.
    """

    def __init__(
        self,
        resolution,
        distance_sensor=None,
        size=None,
        feature_size=None,
        psf_wavelength=[460e-9, 550e-9, 640e-9],
        use_torch=False,
        torch_device="cpu",
        centered=True,
        **kwargs,
    ):
        """
        Constructor from parameters of the user's choice.

        Parameters
        ----------
        resolution: array_like
            Resolution of the  mask (px).
        distance_sensor: float
            Distance between the mask and the sensor (m). Needed to simulate PSF.
        size: array_like
            Size of the mask (m). Only one of ``size`` or ``feature_size`` needs to be specified.
        feature_size: float or array_like
            Size of the feature (m). Only one of ``size`` or ``feature_size`` needs to be specified.
        psf_wavelength: list, optional
            List of wavelengths to simulate PSF (m). Default is [460e-9, 550e-9, 640e-9] nm (blue, green, red).
        use_torch : bool, optional
            If True, the mask is created as a torch tensor. Default is False.
        torch_device : str, optional
            Device to use for torch tensor. Default is 'cpu'.
        centered: bool, optional
            If True, the mask is centered. Default is True.
        """

        resolution = np.array(resolution)
        assert len(resolution) == 2, "Sensor resolution should be of length 2"

        assert (
            size is not None or feature_size is not None
        ), "Either sensor_size or feature_size should be specified"
        if size is None:
            size = np.array(resolution * feature_size)
        else:
            size = np.array(size)
            assert len(size) == 2, "Sensor size should be of length 2"
        if feature_size is None:
            feature_size = np.array(size) / np.array(resolution)
        else:
            if isinstance(feature_size, float):
                feature_size = np.array([feature_size, feature_size])
            else:
                assert len(feature_size) == 2, "Feature size should be of length 2"
                feature_size = np.array(feature_size)
            assert np.all(feature_size > 0), "Feature size should be positive"
        assert np.all(resolution * feature_size <= size)

        self.resolution = resolution
        self.resolution = (int(self.resolution[0]), int(self.resolution[1]))
        self.size = size
        self.centered = centered
        if feature_size is None:
            self.feature_size = self.size / self.resolution
        else:
            self.feature_size = feature_size
        self.distance_sensor = distance_sensor

        if use_torch:
            assert torch_available, "PyTorch is not available"
        self.use_torch = use_torch
        self.torch_device = torch_device

        # create mask
        self.height_map = None  # for phase masks
        self.create_mask()  # creates self.mask
        self.shape = self.height_map.shape if self.height_map is not None else self.mask.shape

        # PSF
        assert hasattr(psf_wavelength, "__len__"), "psf_wavelength should be a list"
        self.psf_wavelength = psf_wavelength
        self.psf = None
        if self.distance_sensor is not None:
            self.compute_psf()

    @classmethod
    def from_sensor(cls, sensor_name, downsample=None, **kwargs):
        """
        Constructor from an existing virtual sensor that copies over the sensor parameters
        (sensor resolution, sensor size, feature size).

        Parameters
        ----------
        sensor_name: str
            Name of the sensor. See :py:class:`~lensless.hardware.sensor.SensorOptions`.
        downsample: float, optional
            Downsampling factor.
        **kwargs:
            Additional arguments for the mask constructor. See the abstract class :py:class:`~lensless.hardware.mask.Mask`
            and the corresponding subclass for more details.

        Example
        -------

        .. code-block:: python

            mask = CodedAperture.from_sensor(sensor_name=SensorOptions.RPI_HQ, downsample=8, ...)
        """
        sensor = VirtualSensor.from_name(sensor_name, downsample)
        return cls(
            resolution=tuple(sensor.resolution.copy()),
            size=tuple(sensor.size.copy()),
            feature_size=sensor.pixel_size.copy(),
            **kwargs,
        )

    @abc.abstractmethod
    def create_mask(self):
        """
        Abstract mask creation method that creates mask with subclass-specific function.
        """
        pass

    def height_map_to_field(self, wavelength, return_phase=False):
        """
        Compute phase from height map.

        Parameters
        ----------
        height_map: :py:class:`~numpy.ndarray`
            Height map.
        wavelength: float
            Wavelength of the light (m).
        return_phase: bool, optional
            If True, return the phase instead of the field. Default is False.
        """
        assert self.height_map is not None, "Height map should be computed first."
        assert self.refractive_index is not None, "Refractive index should be specified."

        phase_pattern = self.height_map * (self.refractive_index - 1) * 2 * np.pi / wavelength
        if return_phase:
            return phase_pattern
        else:
            return (
                np.exp(1j * phase_pattern) if not self.use_torch else torch.exp(1j * phase_pattern)
            )

    def compute_psf(self, distance_sensor=None, wavelength=None, intensity=True):
        """
        Compute the intensity PSF with bandlimited angular spectrum (BLAS) for each wavelength.
        Common to all types of masks.

        Parameters
        ----------
        distance_sensor: float, optional
            Distance between mask and sensor (m). Default is the distance specified at initialization.
        wavelength: float or array_like, optional
            Wavelength(s) to compute the PSF (m). Default is the list of wavelengths specified at initialization.
        """
        if distance_sensor is not None:
            self.distance_sensor = distance_sensor
        assert (
            self.distance_sensor is not None
        ), "Distance between mask and sensor should be specified."

        if wavelength is None:
            wavelength = self.psf_wavelength
        else:
            if not hasattr(wavelength, "__len__"):
                wavelength = [wavelength]

        if self.use_torch:
            psf = torch.zeros(
                tuple(self.resolution) + (len(wavelength),),
                dtype=torch.complex64,
                device=self.torch_device,
            )
        else:
            psf = np.zeros(tuple(self.resolution) + (len(wavelength),), dtype=np.complex64)
        for i, wv in enumerate(wavelength):
            psf[:, :, i] = angular_spectrum(
                u_in=self.mask if self.height_map is None else self.height_map_to_field(wv),
                wv=wv,
                d1=self.feature_size,
                dz=self.distance_sensor,
                dtype=np.float32 if not self.use_torch else torch.float32,
                bandlimit=True,
                device=self.torch_device if self.use_torch else None,
            )[0]

        # intensity PSF
        if intensity:
            self.psf = np.abs(psf) ** 2 if not self.use_torch else torch.abs(psf) ** 2
        else:
            self.psf = psf

        return self.psf

    def plot(self, ax=None, **kwargs):
        """
        Plot the mask.

        Parameters
        ----------
        ax: :py:class:`~matplotlib.axes.Axes`, optional
            Axes to plot the mask on. Default is None.
        **kwargs:
            Additional arguments for the plot function.
        """

        if ax is None:
            _, ax = plt.subplots()

        if self.height_map is not None:
            mask = self.height_map
            title = "Height map"
        else:
            mask = self.mask
            title = "Amplitude mask"
        if self.use_torch:
            mask = mask.cpu().numpy()

        if self.centered:
            extent = (
                -self.size[1] / 2 * 1e3,
                self.size[1] / 2 * 1e3,
                self.size[0] / 2 * 1e3,
                -self.size[0] / 2 * 1e3,
            )
        else:
            extent = (0, self.size[1] * 1e3, self.size[0] * 1e3, 0)

        ax.imshow(mask, extent=extent, cmap="gray", **kwargs)
        ax.set_title(title)
        ax.set_xlabel("[mm]")
        ax.set_ylabel("[mm]")
        return ax


class CodedAperture(Mask):
    """
    Coded aperture mask as in `FlatCam <https://arxiv.org/abs/1509.00116>`_.
    """

    def __init__(self, method="MLS", n_bits=8, **kwargs):
        """
        Coded aperture mask contructor (FlatCam).

        Parameters
        ----------
        method: str
            Pattern generation method (MURA or MLS). Default is ``MLS``.
        n_bits: int, optional
            Number of bits for pattern generation, must be prime for MURA.
            Results in ``2^n - 1``x``2^n - 1`` for MLS.
            Default is 8 (for a 255x255 MLS mask).
        **kwargs:
            The keyword arguments are passed to the parent class :py:class:`~lensless.hardware.mask.Mask`.
        """

        self.row = None
        self.col = None
        self.method = method
        self.n_bits = n_bits

        assert self.method.upper() in ["MURA", "MLS"], "Method should be either 'MLS' or 'MURA'"
        # TODO? use: https://github.com/bpops/codedapertures

        # initialize parameters
        if self.method.upper() == "MURA":
            self.mask = self.generate_mura(self.n_bits)
            self.row = None
            self.col = None
        else:
            seq = max_len_seq(self.n_bits)[0] * 2 - 1
            self.row = seq
            self.col = seq

        if "is_torch" in kwargs and kwargs["is_torch"]:
            torch_device = kwargs["torch_device"] if "torch_device" in kwargs else "cpu"
            if self.row is not None and self.col is not None:
                self.row = torch.from_numpy(self.row).float().to(torch_device)
                self.col = torch.from_numpy(self.col).float().to(torch_device)
            else:
                self.mask = torch.from_numpy(self.mask).float().to(torch_device)

        # needs to be done at the end as it calls create_mask
        super().__init__(**kwargs)

    def create_mask(self, row=None, col=None, mask=None):
        """
        Creating coded aperture mask.
        """

        if mask is not None:
            self.mask = mask
            assert row is None and col is None, "Row and col should not be specified"

        elif row is not None:
            assert col is not None, "Both row and col should be specified"
            self.row = row
            self.col = col

        # output product if necessary
        if self.row is not None:
            if self.use_torch:
                self.mask = torch.outer(self.row, self.col)
                self.mask = torch.round((self.mask + 1) / 2).to(torch.uint8)
            else:
                self.mask = np.outer(self.row, self.col)
                self.mask = np.round((self.mask + 1) / 2).astype(np.uint8)
        assert self.mask is not None, "Mask should be specified"

        # resize to sensor shape
        if np.any(self.resolution != self.mask.shape):

            if self.use_torch:
                self.mask = self.mask.unsqueeze(0).unsqueeze(0)
                self.mask = torch.nn.functional.interpolate(
                    self.mask, size=tuple(self.resolution), mode="nearest"
                ).squeeze()
            else:
                # self.mask = resize(self.mask[:, :, np.newaxis], shape=tuple(self.resolution) + (1,))
                self.mask = resize(
                    self.mask[:, :, np.newaxis],
                    shape=tuple(self.resolution) + (1,),
                    interpolation=cv.INTER_NEAREST,
                ).squeeze()

    def is_prime(self, n):
        """
        Assess whether a number is prime or not.

        Parameters
        ----------
        n: int
            The number we want to check.
        """
        if n % 2 == 0 and n > 2:
            return False
        return all(n % i for i in range(3, int(sqrt(n)) + 1, 2))

    def generate_mura(self, p):
        """
        Generate MURA square pattern.

        Parameters
        ----------
        p: int
            Number of bits.
        """

        if not self.is_prime(p):
            raise ValueError("p is not a valid length. It must be prime.")
        A = np.zeros((p, p), dtype=int)
        q = quadratic_residues(p)
        A[1:, 0] = 1
        for j in range(1, p):
            for i in range(1, p):
                if not ((i - 1 in q) != (j - 1 in q)):
                    A[i, j] = 1
        return A

    def get_conv_matrices(self, img_shape):
        """
        Get theoretical left and right convolution matrices for the separable mask.

        Such that measurement model is given ``P @ img @ Q.T``.

        Parameters
        ----------
        img_shape: tuple
            Shape of the image to being convolved.

        Returns
        -------
        P: :py:class:`~numpy.ndarray`
            Left convolution matrix.
        Q: :py:class:`~numpy.ndarray`
            Right convolution matrix.

        """

        P = circulant(np.resize(self.col, self.resolution[0]))[:, : img_shape[0]]
        Q = circulant(np.resize(self.row, self.resolution[1]))[:, : img_shape[1]]

        return P, Q

    def simulate(self, obj, snr_db=20):
        """
        Simulate the mask measurement of an image. Apply left and right convolution matrices,
        add noise and return the measurement.

        Parameters
        ----------
        obj: :py:class:`~numpy.ndarray`
            Image to simulate.
        snr_db: float, optional
            Signal-to-noise ratio (dB) of the simulated measurement. Default is 20 dB.
        """
        assert len(obj.shape) == 3, "Object should be a 3D array (HxWxC) even if grayscale."

        # Get convolution matrices
        P, Q = self.get_conv_matrices(obj.shape)

        # Convolve image
        n_channels = obj.shape[-1]

        if torch_available and isinstance(obj, torch.Tensor):
            P = torch.from_numpy(P).float()
            Q = torch.from_numpy(Q).float()
            meas = torch.dstack(
                [torch.linalg.multi_dot([P, obj[:, :, c], Q.T]) for c in range(n_channels)]
            ).float()
        else:
            meas = np.dstack([multi_dot([P, obj[:, :, c], Q.T]) for c in range(n_channels)])

        # Add noise
        if snr_db is not None:
            meas = add_shot_noise(meas, snr_db=snr_db)

        if torch_available and isinstance(obj, torch.Tensor):
            meas = meas.to(obj)

        return meas


class MultiLensArray(Mask):
    """
    Multi-lens array mask.
    """

    def __init__(
        self,
        N=None,
        radius=None,
        loc=None,
        refractive_index=1.2,
        seed=0,
        min_height=1e-5,
        radius_range=(1e-4, 4e-4),
        min_separation=1e-4,
        focal_range=None,
        verbose=False,
        **kwargs,
    ):
        """
        Multi-lens array mask constructor.

        Parameters
        ----------
        N: int
            Number of micro-lenses.
        radius: array_like
            Radius of the lenses (m).
        loc: array_like of tuples
            Location of the lenses (m).
        refractive_index: float
            Refractive index of the mask substrate. Default is 1.2.
        seed: int
            Seed for the random number generator. Default is 0.
        min_height: float
            Minimum height of the lenses (m). Default is 1e-3.
        radius_range: array_like
            Range of the radius of the lenses (m). Default is (1e-4, 4e-4) m.
        focal_range: array_like
            Range of the focal length of the lenses (m). Default is None. Overrides the radius_range.
        min_separation: float
            Minimum separation between lenses (m). Default is 1e-4.
        verbose: bool
            If True, print lens placement information. Default is False.
        """
        self.N = N
        self.radius = radius
        self.loc = loc
        self.refractive_index = refractive_index
        self.seed = seed
        self.min_height = min_height
        self.min_separation = min_separation
        self.verbose = verbose

        self.radius_range = radius_range
        if focal_range is not None:
            self.radius_range = [
                focal_range[0] / (self.refractive_index - 1),
                focal_range[1] / (self.refractive_index - 1),
            ]

        super().__init__(**kwargs)

    def check_asserts(self):
        """
        Check the validity of the parameters.

        Generate the locations and radii of the lenses if not specified.
        """
        assert (
            self.radius_range[0] < self.radius_range[1]
        ), "Minimum radius should be smaller than maximum radius"
        if self.radius is not None:
            if self.use_torch:
                assert torch.all(self.radius >= 0)
            else:
                assert np.all(self.radius >= 0)
            assert (
                self.loc is not None
            ), "Location of the lenses should be specified if their radius is specified"
            assert len(self.radius) == len(
                self.loc
            ), "Number of radius should be equal to the number of locations"
            self.N = len(self.radius)
            circles = (
                np.array([(self.loc[i][0], self.loc[i][1], self.radius[i]) for i in range(self.N)])
                if not self.use_torch
                else torch.tensor(
                    [(self.loc[i][0], self.loc[i][1], self.radius[i]) for i in range(self.N)]
                ).to(self.torch_device)
            )
            assert self.no_circle_overlap(circles), "lenses should not overlap"
        else:
            # generate random locations and radii
            assert (
                self.N is not None
            ), "If positions are not specified, the number of lenses should be specified"

            np.random.seed(self.seed)
            self.radius = np.random.uniform(self.radius_range[0], self.radius_range[1], self.N)
            # radius get sorted in descending order
            self.loc, self.radius = self.place_spheres_on_plane(self.radius)
            if self.centered:
                self.loc = self.loc - np.array(self.size) / 2
            if self.use_torch:
                self.radius = torch.tensor(self.radius).to(self.torch_device)
                self.loc = torch.tensor(self.loc).to(self.torch_device)

    def no_circle_overlap(self, circles):
        """
        Check if any circle in the list overlaps with another.

        Parameters
        ----------
        circles: array_like
            List of circles, each represented by a tuple (x, y, r) with location (x, y) and radius r.
        """
        for i in range(len(circles)):
            if self.does_circle_overlap(
                circles[i + 1 :], circles[i][0], circles[i][1], circles[i][2]
            ):
                return False
        return True

    def does_circle_overlap(self, circles, x, y, r):
        """Check if a circle overlaps with any in the list."""
        for (cx, cy, cr) in circles:
            if sqrt((x - cx) ** 2 + (y - cy) ** 2) <= (r + cr + self.min_separation):
                return True, (cx, cy, cr)
        return False

    def place_spheres_on_plane(self, radius, max_attempts=1000):
        """Try to place circles of given radius on a 2D plane."""
        placed_circles = []
        rad_sorted = sorted(radius, reverse=True)  # sort the radius in descending order
        loc = []
        r_placed = []
        for r in rad_sorted:
            placed = False
            for _ in range(max_attempts):
                x = np.random.uniform(r, self.size[1] - r)
                y = np.random.uniform(r, self.size[0] - r)
                if not self.does_circle_overlap(placed_circles, x, y, r):
                    placed_circles.append((x, y, r))
                    loc.append([x, y])
                    r_placed.append(r)
                    placed = True
                    if self.verbose:
                        print(f"Placed circle with rad {r}, and center ({x}, {y})")
                    break
            if not placed:
                if self.verbose:
                    print(f"Failed to place circle with rad {r}")
                continue
        if len(r_placed) < self.N:
            warnings.warn(f"Could not place {self.N - len(r_placed)} lenses")
        return np.array(loc, dtype=np.float32), np.array(r_placed, dtype=np.float32)

    def create_mask(self, loc=None, radius=None):
        """
        Creating multi-lens array mask.

        Parameters
        ----------
        loc: array_like of tuples, optional
            Location of the lenses (m).
        radius: array_like, optional
            Radius of the lenses (m).
        """
        if radius is not None:
            self.radius = radius
        if loc is not None:
            self.loc = loc
        self.check_asserts()

        # convert to pixels (assume same size for x and y)
        locs_pix = self.loc * (1 / self.feature_size[0])
        radius_pix = self.radius * (1 / self.feature_size[0])
        self.height_map = self.create_height_map(radius_pix, locs_pix)

    def create_height_map(self, radius, locs):
        height = (
            np.full((self.resolution[0], self.resolution[1]), self.min_height).astype(np.float32)
            if not self.use_torch
            else torch.full((self.resolution[0], self.resolution[1]), self.min_height).to(
                self.torch_device, dtype=torch.float32
            )
        )
        x = (
            np.arange(self.resolution[0]).astype(np.float32)
            if not self.use_torch
            else torch.arange(self.resolution[0]).to(self.torch_device)
        )
        y = (
            np.arange(self.resolution[1]).astype(np.float32)
            if not self.use_torch
            else torch.arange(self.resolution[1]).to(self.torch_device)
        )
        if self.centered:
            x = x - self.resolution[1] / 2
            y = y - self.resolution[0] / 2
        X, Y = (
            np.meshgrid(x, y, indexing="ij")
            if not self.use_torch
            else torch.meshgrid(x, y, indexing="ij")
        )
        for idx, rad in enumerate(radius):
            contribution = self.lens_contribution(X, Y, rad, locs[idx]) * self.feature_size[0]
            contribution[(X - locs[idx][1]) ** 2 + (Y - locs[idx][0]) ** 2 > rad**2] = 0
            height = height + contribution
        height[height < self.min_height] = self.min_height
        return height

    def lens_contribution(self, x, y, radius, loc):
        return (
            np.sqrt(radius**2 - (x - loc[1]) ** 2 - (y - loc[0]) ** 2)
            if not self.use_torch
            else torch.sqrt(radius**2 - (x - loc[1]) ** 2 - (y - loc[0]) ** 2)
        )

    @property
    def focal_length(self):
        """
        Focal length of the lenses.

        As we have a plano-convex lens: 1/f = (n-1) / R -> f = R / (n-1)
        """
        return self.radius / (self.refractive_index - 1)


class PhaseContour(Mask):
    """
    Phase contour mask as in `PhlatCam <https://ieeexplore.ieee.org/document/9076617>`_.
    """

    def __init__(
        self, noise_period=(16, 16), refractive_index=1.2, n_iter=10, design_wv=532e-9, **kwargs
    ):
        """
        Phase contour mask contructor (PhlatCam).

        Parameters
        ----------
        noise_period: array_like, optional
            Noise period of the Perlin noise (px). Default is (8, 8).
        refractive_index: float, optional
            Refractive index of the mask substrate. Default is 1.2.
        n_iter: int, optional
            Number of iterations for the phase retrieval algorithm. Default is 10.
        design_wv: float, optional
            Wavelength used to design the mask (m). Default is 532e-9, as in the PhlatCam paper.
        **kwargs:
            The keyword arguments are passed to the parent class :py:class:`~lensless.hardware.mask.Mask`.
        """

        self.target_psf = None
        self.phase_pattern = None
        self.height_map = None
        self.noise_period = noise_period
        self.refractive_index = refractive_index
        self.n_iter = n_iter
        self.design_wv = design_wv

        super().__init__(**kwargs)

    def create_mask(self):
        """
        Creating phase contour from edges of Perlin noise.
        """

        # Creating Perlin noise
        proper_dim_1 = (self.resolution[0] // self.noise_period[0]) * self.noise_period[0]
        proper_dim_2 = (self.resolution[1] // self.noise_period[1]) * self.noise_period[1]
        noise = generate_perlin_noise_2d((proper_dim_1, proper_dim_2), self.noise_period)

        # Upscaling to correspond to sensor size
        if np.any(self.resolution != noise.shape):
            noise = resize(noise[:, :, np.newaxis], shape=tuple(self.resolution) + (1,)).squeeze()

        # Edge detection
        binary = np.clip(np.round(np.interp(noise, (-1, 1), (0, 1))), a_min=0, a_max=1)
        self.target_psf = cv.Canny(np.interp(binary, (-1, 1), (0, 255)).astype(np.uint8), 0, 255)

        # Computing mask and height map
        assert (
            self.distance_sensor is not None
        ), "Distance between mask and sensor should be specified."
        _, height_map = phase_retrieval(
            target_psf=self.target_psf,
            wv=self.design_wv,
            d1=self.feature_size,
            dz=self.distance_sensor,
            n=self.refractive_index,
            n_iter=self.n_iter,
            height_map=True,
        )
        self.height_map = height_map


def phase_retrieval(target_psf, wv, d1, dz, n=1.2, n_iter=10, height_map=False, phase_wrap=1):
    """
    Iterative phase retrieval algorithm similar to `PhlatCam <https://ieeexplore.ieee.org/document/9076617>`_,
    using Fresnel propagation.

    Parameters
    ----------
    target_psf: array_like
        Target PSF to optimize the phase mask for.
    wv: float
        Wavelength (m).
    d1: float
        Sample period on the sensor i.e. pixel size (m).
    dz: float
        Propagation distance between the mask and the sensor.
    n: float
        Refractive index of the mask substrate. Default is 1.2.
    n_iter: int
        Number of iterations. Default value is 10.
    phase_wrap : int
        How many multiple of (2*pi) to wrap the phase, to limit heights. Default is 1.
    """
    assert isinstance(phase_wrap, int), "phase_wrap should be an integer"
    M_p = np.sqrt(target_psf)

    if hasattr(d1, "__len__"):
        if d1[0] != d1[1]:
            warnings.warn("Non-square pixel, first dimension taken as feature size.")
        d1 = d1[0]

    for _ in range(n_iter):
        # back propagate from sensor to mask
        M_phi = fresnel_conv(M_p, wv, d1, -dz, dtype=np.float32)[0]
        # constrain amplitude at mask to be unity, i.e. phase pattern
        M_phi = np.exp(1j * np.angle(M_phi))
        # forward propagate from mask to sensor
        M_p = fresnel_conv(M_phi, wv, d1, dz, dtype=np.float32)[0]
        # constrain amplitude to be sqrt(PSF)
        M_p = np.sqrt(target_psf) * np.exp(1j * np.angle(M_p))

    phi = (np.angle(M_phi) + 2 * np.pi) % (2 * np.pi * phase_wrap)

    if height_map:
        return phi, wv * phi / (2 * np.pi * (n - 1))
    else:
        return phi


class FresnelZoneAperture(Mask):
    """
    Fresnel Zone Aperture (FZA) mask as in `this work <https://www.nature.com/articles/s41377-020-0289-9>`_,
    namely binarized cosine function.
    """

    def __init__(self, radius=0.56e-3, **kwargs):
        """
        Fresnel Zone Aperture mask contructor.

        Parameters
        ----------
        radius: float
            Radius of the FZA (m). Default value is 0.56e-3 (largest in the paper, others are 0.32e-3 and 0.25e-3).
        **kwargs:
            The keyword arguments are passed to the parent class :py:class:`~lensless.hardware.mask.Mask`.
        """

        self.radius = radius

        super().__init__(**kwargs)

    def create_mask(self):
        """
        Creating binary Fresnel Zone Aperture mask.
        """
        dim = self.resolution
        x, y = np.meshgrid(
            np.linspace(-dim[1] / 2, dim[1] / 2 - 1, dim[1]),
            np.linspace(-dim[0] / 2, dim[0] / 2 - 1, dim[0]),
        )
        radius_px = self.radius / self.feature_size[0]
        mask = 0.5 * (1 + np.cos(np.pi * (x**2 + y**2) / radius_px**2))
        self.mask = np.round(mask)



"""

Simulation steps

1) Resize image to PSF dimensions while keeping aspect ratio and setting object height to desired value.
2) Convolve each channel of the image with the PSF.
3) Add noise according to desired SNR.

Reconstruct with ADMM and evaluate.

"""

import hydra
from hydra.utils import to_absolute_path
from lensless.utils.io import save_image
from lensless.utils.image import rgb2gray
import numpy as np
import matplotlib.pyplot as plt
from lensless import ADMM
from lensless.utils.plot import plot_image
from lensless.eval.metric import mse, psnr, ssim, lpips
from waveprop.simulation import FarFieldSimulator
import os

from lensless.utils.io import load_image, load_psf


@hydra.main(
    version_base=None, config_path="../../configs/simulate", config_name="simulate_single_file"
)
def simulate(config):

    fp = to_absolute_path(config.files.original)
    assert os.path.exists(fp), f"File {fp} does not exist."
    psf_fp = to_absolute_path(config.files.psf)
    assert os.path.exists(psf_fp), f"PSF {psf_fp} does not exist."

    # simulation parameters
    object_height = config.simulation.object_height
    scene2mask = config.simulation.scene2mask
    mask2sensor = config.simulation.mask2sensor
    sensor = config.simulation.sensor
    snr_db = config.simulation.snr_db
    downsample = config.simulation.downsample
    grayscale = config.simulation.grayscale
    max_val = config.simulation.max_val

    # load image as numpy array
    print("\nImage:")
    image = load_image(fp, verbose=True)
    if grayscale and len(image.shape) == 3:
        image = rgb2gray(image)

    # load psf as numpy array
    print("\nPSF:")
    psf = load_psf(psf_fp, verbose=True, downsample=downsample)
    if grayscale and psf.shape[-1] == 3:
        psf = rgb2gray(psf)
    if downsample > 1:
        print(f"Downsampled to {psf.shape}.")

    """ Simulation"""
    simulator = FarFieldSimulator(
        psf=psf.squeeze(),  # only support one depth plane
        object_height=object_height,
        scene2mask=scene2mask,
        mask2sensor=mask2sensor,
        sensor=sensor,
        snr_db=snr_db,
        max_val=max_val,
    )
    image_plane, object_plane = simulator.propagate(image, return_object_plane=True)
    if config.save:
        save_image(object_plane, "object_plane.png")
        save_image(image_plane, "image_plane.png", max_val=max_val)
    ax = plot_image(image_plane)
    ax.set_title("Simulated lensless image")

    """ Reconstruction """
    recon = ADMM(psf, **config.admm)
    recon.set_data(image_plane[None, :, :, None])
    res = recon.apply(n_iter=config.admm.n_iter, disp_iter=config.admm.disp_iter)
    recovered = res[0]

    """ Evaluation """
    object_plane = object_plane.astype(np.float32)
    recovered = recovered.astype(np.float32).squeeze()

    _, ax = plt.subplots(ncols=2, nrows=1, figsize=(10, 5))
    plot_image(recovered, ax=ax[0])
    ax[0].set_title("Reconstruction")
    plot_image(object_plane, ax=ax[1])
    ax[1].set_title("Original")
    if config.save:
        save_image(recovered, "reconstruction.png")

    print("\nEvaluation:")
    print("MSE", mse(object_plane, recovered))
    print("PSNR", psnr(object_plane, recovered))
    if grayscale:
        print("SSIM", ssim(object_plane, recovered, channel_axis=None))
    else:
        print("SSIM", ssim(object_plane, recovered))
        print("LPIPS", lpips(object_plane, recovered))

    plt.show()


if __name__ == "__main__":
    simulate()



"""
Apply ADMM reconstruction.
```
python scripts/recon/admm.py
```

"""

import hydra
from hydra.utils import to_absolute_path
import os
import time
import pathlib as plib
import matplotlib.pyplot as plt
import numpy as np
from lensless.utils.io import load_data, load_image
from lensless import ADMM
from lensless.utils.plot import plot_image


@hydra.main(version_base=None, config_path="../../configs/recon", config_name="defaults")
def admm(config):
    if config.torch:
        try:
            import torch
        except ImportError:
            raise ImportError("Pytorch not found. Please install pytorch to use torch mode.")

    psf, data = load_data(
        psf_fp=to_absolute_path(config.input.psf),
        data_fp=to_absolute_path(config.input.data),
        background_fp=(
            to_absolute_path(config.input.background)
            if config.input.background is not None
            else None
        ),
        dtype=config.input.dtype,
        downsample=config["preprocess"]["downsample"],
        bayer=config["preprocess"]["bayer"],
        blue_gain=config["preprocess"]["blue_gain"],
        red_gain=config["preprocess"]["red_gain"],
        plot=config["display"]["plot"],
        flip=config["preprocess"]["flip"],
        gamma=config["display"]["gamma"],
        gray=config["preprocess"]["gray"],
        single_psf=config["preprocess"]["single_psf"],
        shape=config["preprocess"]["shape"],
        use_torch=config.torch,
        torch_device=config.torch_device,
        bg_pix=config.preprocess.bg_pix,
        normalize=config.preprocess.normalize,
    )

    disp = config["display"]["disp"]
    if disp < 0:
        disp = None

    save = config["save"]
    if save:
        save = os.getcwd()

    if save:
        if config.torch:
            org_data = data.cpu().numpy()
        else:
            org_data = data
        ax = plot_image(org_data, gamma=config["display"]["gamma"])
        ax.set_title("Raw data")
        plt.savefig(plib.Path(save) / "lensless.png")

        # close axes
        fig = plt.gcf()
        plt.close(fig)

    # load model
    start_time = time.time()
    if not config.admm.unrolled:
        recon = ADMM(psf, **config.admm)
    else:
        assert config.torch, "Unrolled ADMM only works with torch"
        from lensless.recon.unrolled_admm import UnrolledADMM
        import lensless.recon.utils

        pre_process, _ = lensless.recon.utils.create_process_network(
            network=config.admm.pre_process_model.network,
            depth=config.admm.pre_process_model.depth,
            device=config.torch_device,
        )
        post_process, _ = lensless.recon.utils.create_process_network(
            network=config.admm.post_process_model.network,
            depth=config.admm.post_process_model.depth,
            device=config.torch_device,
        )

        recon = UnrolledADMM(psf, pre_process=pre_process, post_process=post_process, **config.admm)
        path = to_absolute_path(config.admm.checkpoint_fp)
        print("Loading checkpoint from : ", path)
        assert os.path.exists(path), "Checkpoint does not exist"
        recon.load_state_dict(torch.load(path, map_location=config.torch_device))

    recon.set_data(data)
    print(f"Setup time : {time.time() - start_time} s")

    start_time = time.time()
    if config.torch:
        with torch.no_grad():
            if config.admm.unrolled:
                res = recon.apply(
                    disp_iter=disp,
                    save=save,
                    gamma=config["display"]["gamma"],
                    plot=config["display"]["plot"],
                    output_intermediate=True,
                )
            else:
                res = recon.apply(
                    disp_iter=disp,
                    save=save,
                    gamma=config["display"]["gamma"],
                    plot=config["display"]["plot"],
                )
    else:
        res = recon.apply(
            disp_iter=disp,
            save=save,
            gamma=config["display"]["gamma"],
            plot=config["display"]["plot"],
        )
    print(f"Processing time : {time.time() - start_time} s")

    if config.torch:
        img = res[0].cpu().numpy()
    else:
        img = res[0]

    if config["display"]["plot"]:
        plt.show()
    if save:

        if config.admm.unrolled:
            # Save intermediate results
            if res[1] is not None:
                pre_processed_image = res[1].cpu().numpy()
                ax = plot_image(pre_processed_image, gamma=config["display"]["gamma"])
                ax.set_title("Image after preprocessing")
                plt.savefig(plib.Path(save) / "pre_processed.png")

            if res[2] is not None:
                pre_post_process_image = res[2].cpu().numpy()
                ax = plot_image(pre_post_process_image, gamma=config["display"]["gamma"])
                ax.set_title("Image prior to post-processing")
                plt.savefig(plib.Path(save) / "pre_post_process.png")

        np.save(plib.Path(save) / "final_reconstruction.npy", img)

        if config.input.original is not None:
            original = load_image(
                to_absolute_path(config.input.original),
                flip=config["preprocess"]["flip"],
                red_gain=config["preprocess"]["red_gain"],
                blue_gain=config["preprocess"]["blue_gain"],
                shape=img.shape[-3:],
            )
            ax = plot_image(original, gamma=config["display"]["gamma"])
            ax.set_title("Ground truth image")
            plt.savefig(plib.Path(save) / "original.png")

            # compute metrics
            from torchmetrics.image import lpip, psnr

            lpips_func = lpip.LearnedPerceptualImagePatchSimilarity(net_type="vgg", normalize=True)
            psnr_funct = psnr.PeakSignalNoiseRatio()

            img_torch = torch.from_numpy(img).squeeze(0)
            original_torch = torch.from_numpy(original).unsqueeze(0)

            # channel as first dimension
            img_torch = img_torch.movedim(-1, -3)
            original_torch = original_torch.movedim(-1, -3)

            # normalize, TODO img max value is 14 which seems strange
            img_torch = img_torch / torch.amax(img_torch)

            # compute metrics
            lpips = lpips_func(img_torch, original_torch)
            psnr = psnr_funct(img_torch, original_torch)
            print(f"LPIPS : {lpips}")
            print(f"PSNR : {psnr}")

        # If the recon algorithm is unrolled and has a preprocessing step, plot result without preprocessing
        if config.admm.unrolled and recon.pre_process is not None:
            recon.set_data(data)
            recon.pre_process = None
            with torch.no_grad():
                res = recon.apply(
                    disp_iter=disp,
                    save=False,
                    gamma=config["display"]["gamma"],
                    plot=config["display"]["plot"],
                    output_intermediate=True,
                )

            img = res[0].cpu().numpy()
            np.save(plib.Path(save) / "final_reconstruction_no_preprocessing.npy", img[0])
            ax = plot_image(img, gamma=config["display"]["gamma"])
            plt.savefig(plib.Path(save) / "final_reconstruction_no_preprocessing.png")
            pre_post_process_image = res[2].cpu().numpy()
            ax = plot_image(pre_post_process_image, gamma=config["display"]["gamma"])
            plt.savefig(plib.Path(save) / "pre_post_process_no_preprocessing.png")

        print(f"Files saved to : {save}")


if __name__ == "__main__":
    admm()



## Adapted from: https://github.com/swz30/Restormer/blob/main/basicsr/models/archs/restormer_arch.py
##
## Restormer: Efficient Transformer for High-Resolution Image Restoration
## Syed Waqas Zamir, Aditya Arora, Salman Khan, Munawar Hayat, Fahad Shahbaz Khan, and Ming-Hsuan Yang
## https://arxiv.org/abs/2111.09881


import torch
import torch.nn as nn
import torch.nn.functional as F
import numbers

from einops import rearrange


def get_restormer_function(model):
    def process_image(image, **kwargs):

        # convert from NDHWC to NCHW
        depth = image.shape[-4]
        image = image.movedim(-1, -3)
        image = image.reshape(-1, *image.shape[-3:])

        # pad to multiple of 8
        img_multiple_of = 8
        h, w = image.shape[-2], image.shape[-1]
        H, W = ((h + img_multiple_of) // img_multiple_of) * img_multiple_of, (
            (w + img_multiple_of) // img_multiple_of
        ) * img_multiple_of
        padh = H - h if h % img_multiple_of != 0 else 0
        padw = W - w if w % img_multiple_of != 0 else 0

        # image = F.pad(image, (0,padw,0,padh), 'reflect')
        image = F.pad(image, (0, padw, 0, padh), mode="constant", value=0)

        # apply model
        image = model(image)
        image = torch.clamp(image, 0, 1)

        # remove padding
        image = image[:, :, :h, :w]

        # convert back to NDHWC
        image = image.movedim(-3, -1)
        image = image.reshape(-1, depth, *image.shape[-3:])

        return image

    return process_image


##########################################################################
## Layer Norm


def to_3d(x):
    return rearrange(x, "b c h w -> b (h w) c")


def to_4d(x, h, w):
    return rearrange(x, "b (h w) c -> b c h w", h=h, w=w)


class BiasFree_LayerNorm(nn.Module):
    def __init__(self, normalized_shape):
        super(BiasFree_LayerNorm, self).__init__()
        if isinstance(normalized_shape, numbers.Integral):
            normalized_shape = (normalized_shape,)
        normalized_shape = torch.Size(normalized_shape)

        assert len(normalized_shape) == 1

        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.normalized_shape = normalized_shape

    def forward(self, x):
        sigma = x.var(-1, keepdim=True, unbiased=False)
        return x / torch.sqrt(sigma + 1e-5) * self.weight


class WithBias_LayerNorm(nn.Module):
    def __init__(self, normalized_shape):
        super(WithBias_LayerNorm, self).__init__()
        if isinstance(normalized_shape, numbers.Integral):
            normalized_shape = (normalized_shape,)
        normalized_shape = torch.Size(normalized_shape)

        assert len(normalized_shape) == 1

        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        self.normalized_shape = normalized_shape

    def forward(self, x):
        mu = x.mean(-1, keepdim=True)
        sigma = x.var(-1, keepdim=True, unbiased=False)
        return (x - mu) / torch.sqrt(sigma + 1e-5) * self.weight + self.bias


class LayerNorm(nn.Module):
    def __init__(self, dim, LayerNorm_type):
        super(LayerNorm, self).__init__()
        if LayerNorm_type == "BiasFree":
            self.body = BiasFree_LayerNorm(dim)
        else:
            self.body = WithBias_LayerNorm(dim)

    def forward(self, x):
        h, w = x.shape[-2:]
        return to_4d(self.body(to_3d(x)), h, w)


##########################################################################
## Gated-Dconv Feed-Forward Network (GDFN)
class FeedForward(nn.Module):
    def __init__(self, dim, ffn_expansion_factor, bias):
        super(FeedForward, self).__init__()

        hidden_features = int(dim * ffn_expansion_factor)

        self.project_in = nn.Conv2d(dim, hidden_features * 2, kernel_size=1, bias=bias)

        self.dwconv = nn.Conv2d(
            hidden_features * 2,
            hidden_features * 2,
            kernel_size=3,
            stride=1,
            padding=1,
            groups=hidden_features * 2,
            bias=bias,
        )

        self.project_out = nn.Conv2d(hidden_features, dim, kernel_size=1, bias=bias)

    def forward(self, x):
        x = self.project_in(x)
        x1, x2 = self.dwconv(x).chunk(2, dim=1)
        x = F.gelu(x1) * x2
        x = self.project_out(x)
        return x


##########################################################################
## Multi-DConv Head Transposed Self-Attention (MDTA)
class Attention(nn.Module):
    def __init__(self, dim, num_heads, bias):
        super(Attention, self).__init__()
        self.num_heads = num_heads
        self.temperature = nn.Parameter(torch.ones(num_heads, 1, 1))

        self.qkv = nn.Conv2d(dim, dim * 3, kernel_size=1, bias=bias)
        self.qkv_dwconv = nn.Conv2d(
            dim * 3, dim * 3, kernel_size=3, stride=1, padding=1, groups=dim * 3, bias=bias
        )
        self.project_out = nn.Conv2d(dim, dim, kernel_size=1, bias=bias)

    def forward(self, x):
        b, c, h, w = x.shape

        qkv = self.qkv_dwconv(self.qkv(x))
        q, k, v = qkv.chunk(3, dim=1)

        # q = rearrange(q, "b (head c) h w -> b head c (h w)", head=self.num_heads)
        # k = rearrange(k, "b (head c) h w -> b head c (h w)", head=self.num_heads)
        # v = rearrange(v, "b (head c) h w -> b head c (h w)", head=self.num_heads)

        q = rearrange(q, "b (head c) h w -> b head c (h w)", head=self.num_heads).contiguous(memory_format=torch.contiguous_format)
        k = rearrange(k, "b (head c) h w -> b head c (h w)", head=self.num_heads).contiguous(memory_format=torch.contiguous_format)
        v = rearrange(v, "b (head c) h w -> b head c (h w)", head=self.num_heads).contiguous(memory_format=torch.contiguous_format)

        q = torch.nn.functional.normalize(q, dim=-1)
        k = torch.nn.functional.normalize(k, dim=-1)

        attn = (q @ k.transpose(-2, -1)) * self.temperature
        attn = attn.softmax(dim=-1)

        out = attn @ v

        out = rearrange(out, "b head c (h w) -> b (head c) h w", head=self.num_heads, h=h, w=w)

        out = self.project_out(out)
        return out


##########################################################################
class TransformerBlock(nn.Module):
    def __init__(self, dim, num_heads, ffn_expansion_factor, bias, LayerNorm_type):
        super(TransformerBlock, self).__init__()

        self.norm1 = LayerNorm(dim, LayerNorm_type)
        self.attn = Attention(dim, num_heads, bias)
        self.norm2 = LayerNorm(dim, LayerNorm_type)
        self.ffn = FeedForward(dim, ffn_expansion_factor, bias)

    def forward(self, x):
        x = x + self.attn(self.norm1(x))
        x = x + self.ffn(self.norm2(x))

        return x


##########################################################################
## Overlapped image patch embedding with 3x3 Conv
class OverlapPatchEmbed(nn.Module):
    def __init__(self, in_c=3, embed_dim=48, bias=False):
        super(OverlapPatchEmbed, self).__init__()

        self.proj = nn.Conv2d(in_c, embed_dim, kernel_size=3, stride=1, padding=1, bias=bias)

    def forward(self, x):
        x = self.proj(x)

        return x


##########################################################################
## Resizing modules
class Downsample(nn.Module):
    def __init__(self, n_feat):
        super(Downsample, self).__init__()

        self.body = nn.Sequential(
            nn.Conv2d(n_feat, n_feat // 2, kernel_size=3, stride=1, padding=1, bias=False),
            nn.PixelUnshuffle(2),
        )

    def forward(self, x):
        return self.body(x)


class Upsample(nn.Module):
    def __init__(self, n_feat):
        super(Upsample, self).__init__()

        self.body = nn.Sequential(
            nn.Conv2d(n_feat, n_feat * 2, kernel_size=3, stride=1, padding=1, bias=False),
            nn.PixelShuffle(2),
        )

    def forward(self, x):
        return self.body(x)


##########################################################################
# ---------- Restormer -----------------------
class Restormer(nn.Module):
    def __init__(
        self,
        inp_channels=3,
        out_channels=3,
        dim=48,
        num_blocks=[4, 6, 6, 8],
        num_refinement_blocks=4,
        heads=[1, 2, 4, 8],
        ffn_expansion_factor=2.66,
        bias=False,
        LayerNorm_type="WithBias",  # Other option 'BiasFree'
        dual_pixel_task=False,  # True for dual-pixel defocus deblurring only. Also set inp_channels=6
    ):

        super(Restormer, self).__init__()

        self.patch_embed = OverlapPatchEmbed(inp_channels, dim)

        self.encoder_level1 = nn.Sequential(
            *[
                TransformerBlock(
                    dim=dim,
                    num_heads=heads[0],
                    ffn_expansion_factor=ffn_expansion_factor,
                    bias=bias,
                    LayerNorm_type=LayerNorm_type,
                )
                for i in range(num_blocks[0])
            ]
        )

        self.down1_2 = Downsample(dim)  # From Level 1 to Level 2
        self.encoder_level2 = nn.Sequential(
            *[
                TransformerBlock(
                    dim=int(dim * 2**1),
                    num_heads=heads[1],
                    ffn_expansion_factor=ffn_expansion_factor,
                    bias=bias,
                    LayerNorm_type=LayerNorm_type,
                )
                for i in range(num_blocks[1])
            ]
        )

        self.down2_3 = Downsample(int(dim * 2**1))  # From Level 2 to Level 3
        self.encoder_level3 = nn.Sequential(
            *[
                TransformerBlock(
                    dim=int(dim * 2**2),
                    num_heads=heads[2],
                    ffn_expansion_factor=ffn_expansion_factor,
                    bias=bias,
                    LayerNorm_type=LayerNorm_type,
                )
                for i in range(num_blocks[2])
            ]
        )

        self.down3_4 = Downsample(int(dim * 2**2))  # From Level 3 to Level 4
        self.latent = nn.Sequential(
            *[
                TransformerBlock(
                    dim=int(dim * 2**3),
                    num_heads=heads[3],
                    ffn_expansion_factor=ffn_expansion_factor,
                    bias=bias,
                    LayerNorm_type=LayerNorm_type,
                )
                for i in range(num_blocks[3])
            ]
        )

        self.up4_3 = Upsample(int(dim * 2**3))  # From Level 4 to Level 3
        self.reduce_chan_level3 = nn.Conv2d(
            int(dim * 2**3), int(dim * 2**2), kernel_size=1, bias=bias
        )
        self.decoder_level3 = nn.Sequential(
            *[
                TransformerBlock(
                    dim=int(dim * 2**2),
                    num_heads=heads[2],
                    ffn_expansion_factor=ffn_expansion_factor,
                    bias=bias,
                    LayerNorm_type=LayerNorm_type,
                )
                for i in range(num_blocks[2])
            ]
        )

        self.up3_2 = Upsample(int(dim * 2**2))  # From Level 3 to Level 2
        self.reduce_chan_level2 = nn.Conv2d(
            int(dim * 2**2), int(dim * 2**1), kernel_size=1, bias=bias
        )
        self.decoder_level2 = nn.Sequential(
            *[
                TransformerBlock(
                    dim=int(dim * 2**1),
                    num_heads=heads[1],
                    ffn_expansion_factor=ffn_expansion_factor,
                    bias=bias,
                    LayerNorm_type=LayerNorm_type,
                )
                for i in range(num_blocks[1])
            ]
        )

        self.up2_1 = Upsample(
            int(dim * 2**1)
        )  # From Level 2 to Level 1  (NO 1x1 conv to reduce channels)

        self.decoder_level1 = nn.Sequential(
            *[
                TransformerBlock(
                    dim=int(dim * 2**1),
                    num_heads=heads[0],
                    ffn_expansion_factor=ffn_expansion_factor,
                    bias=bias,
                    LayerNorm_type=LayerNorm_type,
                )
                for i in range(num_blocks[0])
            ]
        )

        self.refinement = nn.Sequential(
            *[
                TransformerBlock(
                    dim=int(dim * 2**1),
                    num_heads=heads[0],
                    ffn_expansion_factor=ffn_expansion_factor,
                    bias=bias,
                    LayerNorm_type=LayerNorm_type,
                )
                for i in range(num_refinement_blocks)
            ]
        )

        #### For Dual-Pixel Defocus Deblurring Task ####
        self.dual_pixel_task = dual_pixel_task
        if self.dual_pixel_task:
            self.skip_conv = nn.Conv2d(dim, int(dim * 2**1), kernel_size=1, bias=bias)
        ###########################

        self.output = nn.Conv2d(
            int(dim * 2**1), out_channels, kernel_size=3, stride=1, padding=1, bias=bias
        )

    def forward(self, inp_img):

        inp_enc_level1 = self.patch_embed(inp_img)
        out_enc_level1 = self.encoder_level1(inp_enc_level1)

        inp_enc_level2 = self.down1_2(out_enc_level1)
        out_enc_level2 = self.encoder_level2(inp_enc_level2)

        inp_enc_level3 = self.down2_3(out_enc_level2)
        out_enc_level3 = self.encoder_level3(inp_enc_level3)

        inp_enc_level4 = self.down3_4(out_enc_level3)
        latent = self.latent(inp_enc_level4)

        inp_dec_level3 = self.up4_3(latent)
        inp_dec_level3 = torch.cat([inp_dec_level3, out_enc_level3], 1)
        inp_dec_level3 = self.reduce_chan_level3(inp_dec_level3)
        out_dec_level3 = self.decoder_level3(inp_dec_level3)

        inp_dec_level2 = self.up3_2(out_dec_level3)
        inp_dec_level2 = torch.cat([inp_dec_level2, out_enc_level2], 1)
        inp_dec_level2 = self.reduce_chan_level2(inp_dec_level2)
        out_dec_level2 = self.decoder_level2(inp_dec_level2)

        inp_dec_level1 = self.up2_1(out_dec_level2)
        inp_dec_level1 = torch.cat([inp_dec_level1, out_enc_level1], 1)
        out_dec_level1 = self.decoder_level1(inp_dec_level1)

        out_dec_level1 = self.refinement(out_dec_level1)

        #### For Dual-Pixel Defocus Deblurring Task ####
        if self.dual_pixel_task:
            out_dec_level1 = out_dec_level1 + self.skip_conv(inp_enc_level1)
            out_dec_level1 = self.output(out_dec_level1)
        ###########################
        else:
            out_dec_level1 = self.output(out_dec_level1) + inp_img

        return out_dec_level1



import csv
import os
import socket
import subprocess
import time
import paramiko
from pprint import pprint
from paramiko.ssh_exception import AuthenticationException, BadHostKeyException, SSHException
from lensless.hardware.sensor import SensorOptions
import cv2
from lensless.utils.image import print_image_info
from lensless.utils.io import load_image
import logging

logging.getLogger("paramiko").setLevel(logging.WARNING)

if os.name == "nt":
    NULL_FILE = "nul"
else:
    NULL_FILE = "/dev/null 2>&1"


def capture(
    rpi_username,
    rpi_hostname,
    sensor,
    bayer,
    exp,
    fn="capture",
    iso=100,
    config_pause=2,
    sensor_mode="0",
    nbits_out=12,
    legacy=True,
    rgb=False,
    gray=False,
    nbits=12,
    down=None,
    awb_gains=None,
    rpi_python="~/LenslessPiCam/lensless_env/bin/python",
    capture_script="~/LenslessPiCam/scripts/measure/on_device_capture.py",
    verbose=False,
    output_path=None,
    **kwargs,
):
    """
    Capture image.

    Parameters
    ----------
    fn : str
        File name captured image.
    rpi_username : str
        Username of Raspberry Pi.
    rpi_hostname : str
        Hostname of Raspberry Pi.
    sensor : str
        Sensor name
    bayer : bool
        Whether to return bayer data (larger file size to transfer back).
    exp : int
        Exposure time in microseconds.
    iso : int
        ISO.
    config_pause : int
        Time to pause after configuring camera.
    sensor_mode : str
        Sensor mode.
    nbits_out : int
        Number of bits of output image.
    legacy : bool
        Whether to use legacy capture software of Raspberry Pi.
    rgb : bool
        Whether to capture RGB image.
    gray : bool
        Whether to capture grayscale image.
    nbits : int
        Number of bits of image.
    down : int
        Downsample factor.
    awb_gains : list
        AWB gains (red, blue).
    rpi_python : str
        Path to Python on Raspberry Pi.
    capture_script : str
        Path to capture script on Raspberry Pi.
    output_path : str
        Path to save image.
    verbose : bool
        Whether to print extra info.

    """

    # check_username_hostname(rpi_username, rpi_hostname)
    assert sensor in SensorOptions.values(), f"Sensor must be one of {SensorOptions.values()}"

    # form command
    remote_fn = "remote_capture"
    pic_command = (
        f"{rpi_python} {capture_script} sensor={sensor} bayer={bayer} fn={remote_fn} exp={exp} iso={iso} "
        f"config_pause={config_pause} sensor_mode={sensor_mode} nbits_out={nbits_out} "
        f"legacy={legacy} rgb={rgb} gray={gray} "
    )
    if nbits > 8:
        pic_command += " sixteen=True"
    if down:
        pic_command += f" down={down}"
    if awb_gains:
        pic_command += f" awb_gains=[{awb_gains[0]},{awb_gains[1]}]"

    if verbose:
        print(f"COMMAND : {pic_command}")

    # take picture
    ssh = subprocess.Popen(
        ["ssh", "%s@%s" % (rpi_username, rpi_hostname), pic_command],
        shell=False,
        # stdout=DEVNULL,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    result = ssh.stdout.readlines()
    error = ssh.stderr.readlines()

    if error != [] and legacy:  # new camera software seems to return error even if it works
        print("ERROR: %s" % error)
        return
    if result == []:
        error = ssh.stderr.readlines()
        print("ERROR: %s" % error)
        return
    else:
        result = [res.decode("UTF-8") for res in result]
        result = [res for res in result if len(res) > 3]
        result_dict = dict()
        for res in result:
            _key = res.split(":")[0].strip()
            _val = "".join(res.split(":")[1:]).strip()
            result_dict[_key] = _val
        # result_dict = dict(map(lambda s: map(str.strip, s.split(":")), result))
        if verbose:
            print("COMMAND OUTPUT : ")
            pprint(result_dict)

    # copy over file
    if (
        "RPi distribution" in result_dict.keys()
        and "bullseye" in result_dict["RPi distribution"]
        and not legacy
    ):

        if bayer:

            # copy over DNG file
            remotefile = f"~/{remote_fn}.dng"
            localfile = f"{fn}.dng"
            if output_path is not None:
                localfile = os.path.join(output_path, localfile)
            if verbose:
                print(f"\nCopying over picture as {localfile}...")
            os.system(
                'scp "%s@%s:%s" %s >%s'
                % (rpi_username, rpi_hostname, remotefile, localfile, NULL_FILE)
            )

            img = load_image(localfile, verbose=True, bayer=bayer, nbits_out=nbits_out)

            # print image properties
            print_image_info(img)

            # save as PNG
            png_out = f"{fn}.png"
            print(f"Saving RGB file as: {png_out}")
            cv2.imwrite(png_out, cv2.cvtColor(img, cv2.COLOR_RGB2BGR))

        else:

            remotefile = f"~/{remote_fn}.png"
            localfile = f"{fn}.png"
            if output_path is not None:
                localfile = os.path.join(output_path, localfile)
            if verbose:
                print(f"\nCopying over picture as {localfile}...")
            os.system(
                'scp "%s@%s:%s" %s >%s'
                % (rpi_username, rpi_hostname, remotefile, localfile, NULL_FILE)
            )

            img = load_image(localfile, verbose=True)

    # legacy software running on RPi
    else:
        # copy over file
        # more pythonic? https://stackoverflow.com/questions/250283/how-to-scp-in-python
        remotefile = f"~/{remote_fn}.png"
        localfile = f"{fn}.png"
        if output_path is not None:
            localfile = os.path.join(output_path, localfile)
        if verbose:
            print(f"\nCopying over picture as {localfile}...")
        os.system(
            'scp "%s@%s:%s" %s >%s' % (rpi_username, rpi_hostname, remotefile, localfile, NULL_FILE)
        )

        if rgb or gray:
            img = load_image(localfile, verbose=verbose)

        else:

            if not bayer:
                # red_gain = config.camera.red_gain
                # blue_gain = config.camera.blue_gain
                red_gain = awb_gains[0]
                blue_gain = awb_gains[1]
            else:
                red_gain = None
                blue_gain = None

            # load image
            if verbose:
                print("\nLoading picture...")

            img = load_image(
                localfile,
                verbose=True,
                bayer=bayer,
                blue_gain=blue_gain,
                red_gain=red_gain,
                nbits_out=nbits_out,
            )

            # write RGB data
            if not bayer:
                cv2.imwrite(localfile, cv2.cvtColor(img, cv2.COLOR_RGB2BGR))

    return localfile, img


def display(
    fp,
    rpi_username,
    rpi_hostname,
    screen_res,
    image_res=None,
    brightness=100,
    rot90=0,
    pad=0,
    vshift=0,
    hshift=0,
    verbose=False,
    **kwargs,
):
    """
    Display image on a screen.

    Assumes setup described here: https://lensless.readthedocs.io/en/latest/measurement.html#remote-display

    Parameters
    ----------
    fp : str
        File path to image.
    rpi_username : str
        Username of Raspberry Pi.
    rpi_hostname : str
        Hostname of Raspberry Pi.
    screen_res : tuple
        Screen resolution of Raspberry Pi.
    """

    # assumes that `LenslessPiCam` is in home directory and environment inside `LenslessPiCam`
    rpi_python = "~/LenslessPiCam/lensless_env/bin/python"
    script = "~/LenslessPiCam/scripts/measure/prep_display_image.py"
    remote_tmp_file = "~/tmp_display.png"
    display_path = "~/LenslessPiCam_display/test.png"

    os.system(f"scp {fp} {rpi_username}@{rpi_hostname}:{remote_tmp_file} > {NULL_FILE}")

    # run script on Raspberry Pi to prepare image to display
    prep_command = f"{rpi_python} {script} --fp {remote_tmp_file} \
        --pad {pad} --vshift {vshift} --hshift {hshift} --screen_res {screen_res[0]} {screen_res[1]} \
        --brightness {brightness} --rot90 {rot90} --output_path {display_path} "
    if image_res is not None:
        prep_command += f" --image_res {image_res[0]} {image_res[1]}"
    if verbose:
        print(f"COMMAND : {prep_command}")
    subprocess.Popen(
        ["ssh", "%s@%s" % (rpi_username, rpi_hostname), prep_command],
        shell=False,
        # stdout=DEVNULL
    )


def check_username_hostname(username, hostname, timeout=10):

    assert username is not None, "Raspberry Pi username must be specified"
    assert hostname is not None, "Raspberry Pi hostname must be specified"

    client = paramiko.client.SSHClient()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    try:
        # with suppress_stdout():
        client.connect(hostname, username=username, timeout=timeout)
    except (BadHostKeyException, AuthenticationException, SSHException, socket.error) as e:
        raise ValueError(f"Could not connect to {username}@{hostname}\n{e}")

    return client


def get_distro():
    """
    Get current OS distribution.
    Returns
    -------
    result : str
        Name and version of OS.
    """
    # https://majornetwork.net/2019/11/get-linux-distribution-name-and-version-with-python/
    RELEASE_DATA = {}
    with open("/etc/os-release") as f:
        reader = csv.reader(f, delimiter="=")
        for row in reader:
            if row:
                RELEASE_DATA[row[0]] = row[1]
    if RELEASE_DATA["ID"] in ["debian", "raspbian"]:
        with open("/etc/debian_version") as f:
            DEBIAN_VERSION = f.readline().strip()
        major_version = DEBIAN_VERSION.split(".")[0]
        version_split = RELEASE_DATA["VERSION"].split(" ", maxsplit=1)
        if version_split[0] == major_version:
            # Just major version shown, replace it with the full version
            RELEASE_DATA["VERSION"] = " ".join([DEBIAN_VERSION] + version_split[1:])
    return f"{RELEASE_DATA['NAME']} {RELEASE_DATA['VERSION']}"


def set_mask_sensor_distance(
    distance, rpi_username, rpi_hostname, motor=1, max_distance=16, timeout=5
):
    """
    Set the distance between the mask and sensor.

    This functions assumes that `StepperDriver <https://github.com/adrienhoffet/StepperDriver>`_ is installed.
    is downloaded on the Raspberry Pi.

    Parameters
    ----------
    distance : float
        Distance in mm. Positive values move the mask away from the sensor.
    rpi_username : str
        Username of Raspberry Pi.
    rpi_hostname : str
        Hostname of Raspberry Pi.
    """

    client = check_username_hostname(rpi_username, rpi_hostname)
    assert motor in [0, 1]
    assert distance >= 0, "Distance must be non-negative"
    assert distance <= max_distance, f"Distance must be less than {max_distance} mm"

    # assumes that `StepperDriver` is in home directory
    rpi_python = "python3"
    script = "~/StepperDriver/Python/serial_motors.py"

    # reset to zero
    print("Resetting to zero distance...")
    try:
        command = f"{rpi_python} {script} {motor} REV {max_distance * 1000}"
        _stdin, _stdout, _stderr = client.exec_command(command, timeout=timeout)
    except socket.timeout:  # socket.timeout
        pass

    client.close()
    time.sleep(5)  # TODO reduce this time
    client = check_username_hostname(rpi_username, rpi_hostname)

    # set to desired distance
    if distance != 0:
        print(f"Setting distance to {distance} mm...")
        distance_um = distance * 1000
        if distance_um >= 0:
            command = f"{rpi_python} {script} {motor} FWD {distance_um}"
        else:
            command = f"{rpi_python} {script} {motor} REV {-1 * distance_um}"
        print(f"COMMAND : {command}")
        try:
            _stdin, _stdout, _stderr = client.exec_command(command, timeout=timeout)
            print(_stdout.read().decode())
        except socket.timeout:  # socket.timeout
            client.close()

    client.close()



from diffusercam_admm import apply_admm
from lensless.recon.admm import ADMM
from lensless.utils.io import load_data
import time
import pathlib as plib
import numpy as np


psf_fp = "data/psf/tape_rgb.png"
data_fp = "data/raw_data/thumbs_up_rgb.png"

n_iter = 5
downsample = 4
gray = True
dtype = "float32"
n_trials = 10


psf, data = load_data(
    psf_fp=psf_fp,
    data_fp=data_fp,
    downsample=downsample,
    plot=False,
    gray=gray,
    dtype=dtype,
)

save = "profile_admm"
save = plib.Path(__file__).parent / save
save.mkdir(exist_ok=True)

""" LenslessPiCam """
recon = ADMM(psf, dtype=dtype)
recon.set_data(data)
res = recon.apply(n_iter=n_iter, save=save, disp_iter=None, plot=False)
recon.reset()
total_time = 0
for _ in range(n_trials):
    start_time = time.time()
    res = recon.apply(n_iter=n_iter, disp_iter=None, plot=False)
    total_time += time.time() - start_time
    recon.reset()
print(f"LenslessPiCam (avg) : {total_time / n_trials} s")

""" DiffuserCam"""
sensor_size = np.array(psf.shape[:2])
full_size = 2 * sensor_size
param = {
    "mu1": 1e-6,
    "mu2": 1e-5,
    "mu3": 4e-5,
    "tau": 0.0001,
    "sensor_size": sensor_size,
    "full_size": full_size,
}
apply_admm(psf, data, n_iter=n_iter, param=param, disp_iter=None, save=save)
start_time = time.time()
for _ in range(n_trials):
    apply_admm(psf, data, n_iter=n_iter, param=param, disp_iter=None)
print(f"DiffuserCam (avg)  : {(time.time() - start_time) / n_trials} s")


""" PyTorch CPU """
psf, data = load_data(
    psf_fp=psf_fp,
    data_fp=data_fp,
    downsample=downsample,
    plot=False,
    gray=gray,
    dtype=dtype,
    use_torch=True,
)

recon = ADMM(psf, dtype=dtype)
recon.set_data(data)
res = recon.apply(n_iter=n_iter, save=save, disp_iter=None, plot=False)
recon.reset()
total_time = 0
for _ in range(n_trials):
    start_time = time.time()
    res = recon.apply(n_iter=n_iter, disp_iter=None, plot=False)
    total_time += time.time() - start_time
    recon.reset()
print(f"LenslessPiCam, PyTorch CPU (avg) : {total_time / n_trials} s")

""" PyTorch GPU """
psf, data = load_data(
    psf_fp=psf_fp,
    data_fp=data_fp,
    downsample=downsample,
    plot=False,
    gray=gray,
    dtype=dtype,
    use_torch=True,
    torch_device="cuda",
)

recon = ADMM(psf, dtype=dtype)
recon.set_data(data)
res = recon.apply(n_iter=n_iter, save=save, disp_iter=None, plot=False)
recon.reset()
total_time = 0
for _ in range(n_trials):
    start_time = time.time()
    res = recon.apply(n_iter=n_iter, disp_iter=None, plot=False)
    total_time += time.time() - start_time
    recon.reset()
print(f"LenslessPiCam, PyTorch GPU (avg) : {total_time / n_trials} s")



# #############################################################################
# network_unet.py
# =================
# Adapted from:
# https://github.com/cszn/DPIR/blob/15bca3fcc1f3cc51a1f99ccf027691e278c19354/models/network_unet.py
# #############################################################################


import torch
import torch.nn as nn
import lensless.recon.drunet.basicblock as B
from torchvision.transforms.functional import resize

"""
# ====================
# unet
# From https://github.com/cszn/DPIR/blob/master/main_dpir_denoising.py
# ====================
"""


class UNet(nn.Module):
    def __init__(
        self,
        in_nc=1,
        out_nc=1,
        nc=[64, 128, 256, 512],
        nb=2,
        act_mode="R",
        downsample_mode="strideconv",
        upsample_mode="convtranspose",
    ):
        super(UNet, self).__init__()

        self.m_head = B.conv(in_nc, nc[0], mode="C" + act_mode[-1])

        # downsample
        if downsample_mode == "avgpool":
            downsample_block = B.downsample_avgpool
        elif downsample_mode == "maxpool":
            downsample_block = B.downsample_maxpool
        elif downsample_mode == "strideconv":
            downsample_block = B.downsample_strideconv
        else:
            raise NotImplementedError("downsample mode [{:s}] is not found".format(downsample_mode))

        self.m_down1 = B.sequential(
            *[B.conv(nc[0], nc[0], mode="C" + act_mode) for _ in range(nb)],
            downsample_block(nc[0], nc[1], mode="2" + act_mode)
        )
        self.m_down2 = B.sequential(
            *[B.conv(nc[1], nc[1], mode="C" + act_mode) for _ in range(nb)],
            downsample_block(nc[1], nc[2], mode="2" + act_mode)
        )
        self.m_down3 = B.sequential(
            *[B.conv(nc[2], nc[2], mode="C" + act_mode) for _ in range(nb)],
            downsample_block(nc[2], nc[3], mode="2" + act_mode)
        )

        self.m_body = B.sequential(
            *[B.conv(nc[3], nc[3], mode="C" + act_mode) for _ in range(nb + 1)]
        )

        # upsample
        if upsample_mode == "upconv":
            upsample_block = B.upsample_upconv
        elif upsample_mode == "pixelshuffle":
            upsample_block = B.upsample_pixelshuffle
        elif upsample_mode == "convtranspose":
            upsample_block = B.upsample_convtranspose
        else:
            raise NotImplementedError("upsample mode [{:s}] is not found".format(upsample_mode))

        self.m_up3 = B.sequential(
            upsample_block(nc[3], nc[2], mode="2" + act_mode),
            *[B.conv(nc[2], nc[2], mode="C" + act_mode) for _ in range(nb)]
        )
        self.m_up2 = B.sequential(
            upsample_block(nc[2], nc[1], mode="2" + act_mode),
            *[B.conv(nc[1], nc[1], mode="C" + act_mode) for _ in range(nb)]
        )
        self.m_up1 = B.sequential(
            upsample_block(nc[1], nc[0], mode="2" + act_mode),
            *[B.conv(nc[0], nc[0], mode="C" + act_mode) for _ in range(nb)]
        )

        self.m_tail = B.conv(nc[0], out_nc, bias=True, mode="C")

    def forward(self, x0):
        x1 = self.m_head(x0)
        x2 = self.m_down1(x1)
        x3 = self.m_down2(x2)
        x4 = self.m_down3(x3)
        x = self.m_body(x4)
        x = self.m_up3(x + x4)
        x = self.m_up2(x + x3)
        x = self.m_up1(x + x2)
        x = self.m_tail(x + x1) + x0

        return x


class UNetRes(nn.Module):
    def __init__(
        self,
        in_nc=1,
        out_nc=1,
        nc=[64, 128, 256, 512],
        nb=4,
        act_mode="R",
        downsample_mode="strideconv",
        upsample_mode="convtranspose",
        concatenate_compensation=False,
        input_background=False,
        background_subtraction=False,
    ):
        super(UNetRes, self).__init__()

        assert len(nc) == 4, "nc's length should be 4."

        self.nc = nc
        self.m_head = B.conv(in_nc, nc[0], bias=False, mode="C")

        # downsample
        if downsample_mode == "avgpool":
            downsample_block = B.downsample_avgpool
        elif downsample_mode == "maxpool":
            downsample_block = B.downsample_maxpool
        elif downsample_mode == "strideconv":
            downsample_block = B.downsample_strideconv
        else:
            raise NotImplementedError("downsample mode [{:s}] is not found".format(downsample_mode))

        self.m_down1 = B.sequential(
            *[B.ResBlock(nc[0], nc[0], bias=False, mode="C" + act_mode + "C") for _ in range(nb)],
            downsample_block(nc[0], nc[1], bias=False, mode="2")
        )
        self.m_down2 = B.sequential(
            *[B.ResBlock(nc[1], nc[1], bias=False, mode="C" + act_mode + "C") for _ in range(nb)],
            downsample_block(nc[1], nc[2], bias=False, mode="2")
        )
        self.m_down3 = B.sequential(
            *[B.ResBlock(nc[2], nc[2], bias=False, mode="C" + act_mode + "C") for _ in range(nb)],
            downsample_block(nc[2], nc[3], bias=False, mode="2")
        )

        # define background subtraction branch (pre-processing)
        self.input_background = input_background
        self.background_subtraction = background_subtraction
        if background_subtraction:
            self.subtraction_weights = nn.Parameter(torch.ones(len(nc)))
            # TODO create downsample block for background
            self.m_head_background = B.conv(in_nc, nc[0], bias=False, mode="C")
            self.m_down1_background = B.sequential(
                *[
                    B.ResBlock(nc[0], nc[0], bias=False, mode="C" + act_mode + "C")
                    for _ in range(nb)
                ],
                downsample_block(nc[0], nc[1], bias=False, mode="2")
            )
            self.m_down2_background = B.sequential(
                *[
                    B.ResBlock(nc[1], nc[1], bias=False, mode="C" + act_mode + "C")
                    for _ in range(nb)
                ],
                downsample_block(nc[1], nc[2], bias=False, mode="2")
            )
            self.m_down3_background = B.sequential(
                *[
                    B.ResBlock(nc[2], nc[2], bias=False, mode="C" + act_mode + "C")
                    for _ in range(nb)
                ],
                downsample_block(nc[2], nc[3], bias=False, mode="2")
            )

        # define compensation branch (post-processing)
        self.concatenate_compensation = concatenate_compensation
        if concatenate_compensation:
            self.m_body = B.sequential(
                B.conv(nc[3] + concatenate_compensation, nc[3], bias=False, mode="C" + act_mode),
                *[
                    B.ResBlock(nc[3], nc[3], bias=False, mode="C" + act_mode + "C")
                    for _ in range(nb)
                ]
            )
        else:
            self.m_body = B.sequential(
                *[
                    B.ResBlock(nc[3], nc[3], bias=False, mode="C" + act_mode + "C")
                    for _ in range(nb)
                ]
            )

        # upsample
        if upsample_mode == "upconv":
            upsample_block = B.upsample_upconv
        elif upsample_mode == "pixelshuffle":
            upsample_block = B.upsample_pixelshuffle
        elif upsample_mode == "convtranspose":
            upsample_block = B.upsample_convtranspose
        else:
            raise NotImplementedError("upsample mode [{:s}] is not found".format(upsample_mode))

        self.m_up3 = B.sequential(
            upsample_block(nc[3], nc[2], bias=False, mode="2"),
            *[B.ResBlock(nc[2], nc[2], bias=False, mode="C" + act_mode + "C") for _ in range(nb)]
        )
        self.m_up2 = B.sequential(
            upsample_block(nc[2], nc[1], bias=False, mode="2"),
            *[B.ResBlock(nc[1], nc[1], bias=False, mode="C" + act_mode + "C") for _ in range(nb)]
        )
        self.m_up1 = B.sequential(
            upsample_block(nc[1], nc[0], bias=False, mode="2"),
            *[B.ResBlock(nc[0], nc[0], bias=False, mode="C" + act_mode + "C") for _ in range(nb)]
        )

        self.m_tail = B.conv(nc[0], out_nc, bias=False, mode="C")

    def forward(self, x0, compensation_output=None, background=None):
        if self.concatenate_compensation:
            assert compensation_output is not None, "compensation_output should not be None."
        if self.background_subtraction:
            assert background is not None, "background should not be None."

        x1 = self.m_head(x0)
        x2 = self.m_down1(x1)
        x3 = self.m_down2(x2)
        x4 = self.m_down3(x3)

        if self.background_subtraction:
            # downsample background
            x1_background = self.m_head_background(background)
            x2_background = self.m_down1_background(x1_background)
            x3_background = self.m_down2_background(x2_background)
            x4_background = self.m_down3_background(x3_background)

            # background subtraction
            x1 = x1 - self.subtraction_weights[0] * x1_background
            x2 = x2 - self.subtraction_weights[1] * x2_background
            x3 = x3 - self.subtraction_weights[2] * x3_background
            x4 = x4 - self.subtraction_weights[3] * x4_background

        if compensation_output is not None:
            compensation_output_re = resize(compensation_output, tuple(x4.shape[-2:]))
            latent = torch.cat([x4, compensation_output_re], dim=1)
        else:
            latent = x4

        x = self.m_body(latent)
        x = self.m_up3(x + x4)
        x = self.m_up2(x + x3)
        x = self.m_up1(x + x2)
        x = self.m_tail(x + x1)

        return x



# #############################################################################
# multi_wiener.py
# ===============
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# Kyung Chul Lee
# #############################################################################


import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from lensless.recon.utils import convert_to_NCHW, convert_to_NDCHW
from lensless.recon.rfft_convolve import RealFFTConvolve2D


class DoubleConv(nn.Module):
    """(convolution => [BN] => ReLU) * 2"""

    def __init__(self, in_channels, out_channels, mid_channels=None):
        super().__init__()
        if not mid_channels:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=(3, 3), padding=1, bias=False),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(mid_channels, out_channels, kernel_size=(3, 3), padding=1, bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        return self.double_conv(x)


class Down(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()
        self.pool_conv = nn.Sequential(
            # nn.AvgPool2d(2),
            nn.MaxPool2d(2),  # original paper says max-pooling
            DoubleConv(in_channels, out_channels),
        )

    def forward(self, x):
        return self.pool_conv(x)


class Up(nn.Module):
    def __init__(self, in_channels, out_channels):
        super().__init__()

        # or use ConvTranspose2d? https://github.com/milesial/Pytorch-UNet/blob/21d7850f2af30a9695bbeea75f3136aa538cfc4a/unet/unet_parts.py#L53
        self.up = nn.Upsample(scale_factor=2, mode="bilinear", align_corners=True)
        self.conv = DoubleConv(in_channels, out_channels, in_channels // 2)

    def forward(self, x1, x2):
        x1 = self.up(x1)
        # input is CHW
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2, diffY // 2, diffY - diffY // 2])
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)


class OutConv(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(OutConv, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=(1, 1))

    def forward(self, x):
        return self.conv(x)


def WieNer(blur, psf, delta):
    blur_fft = torch.fft.rfft2(blur)
    psf_fft = torch.fft.rfft2(psf)
    psf_fft = torch.conj(psf_fft) / (torch.abs(psf_fft) ** 2 + delta)
    img = torch.fft.ifftshift(torch.fft.irfft2(psf_fft * blur_fft), (-2, -1))
    return img.real


class MultiWiener(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        psf,
        psf_channels=1,
        nc=None,
        pre_process=None,
        skip_pre=False,
    ):
        """
        Constructor for Multi-Wiener Deconvolution Network (MWDN) as proposed in:
        https://opg.optica.org/oe/fulltext.cfm?uri=oe-31-23-39088&id=541387

        Parameters
        ----------
        in_channels : int
            Number of input channels. RGB or grayscale, i.e. 3 and 1 respectively.
        out_channels : int
            Number of output channels. RGB or grayscale, i.e. 3 and 1 respectively.
        psf : :py:class:`~torch.Tensor`
            Point spread function (PSF) that models forward propagation.
        psf_channels : int
            Number of channels in the PSF. Default is 1.
        nc : list
            Number of channels in the network. Default is [64, 128, 256, 512, 512].
        pre_process : :py:class:`function` or :py:class:`~torch.nn.Module`, optional
            Pre-processor applies before MWDN. Default is None.
        skip_pre : bool
            Skip pre-processing. Default is False.

        """
        assert in_channels == 1 or in_channels == 3, "in_channels must be 1 or 3"
        assert out_channels == 1 or out_channels == 3, "out_channels must be 1 or 3"
        assert in_channels >= out_channels
        if nc is None:
            nc = [64, 128, 256, 512, 512]

        super(MultiWiener, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels

        self.inc = DoubleConv(in_channels, nc[0])
        self.down_layers = nn.ModuleList([Down(nc[i], nc[i + 1]) for i in range(len(nc) - 1)])

        self.up_layers = []
        n_prev = nc[-1]
        for i in range(len(nc) - 1):
            n_in = n_prev + nc[-i - 2]
            n_out = nc[-i - 2] // 2 if i < len(nc) - 2 else nc[0]
            self.up_layers.append(Up(n_in, n_out))
            n_prev = n_out
        self.up_layers = nn.ModuleList(self.up_layers)
        self.outc = OutConv(nc[0], out_channels)

        self.delta = nn.Parameter(torch.tensor(np.ones(len(nc)) * 0.01, dtype=torch.float32))
        self.w = nn.Parameter(
            torch.tensor(np.ones((1, psf_channels, 1, 1)) * 0.001, dtype=torch.float32)
        )

        self.inc0 = DoubleConv(psf_channels, nc[0])
        self.psf_down = nn.ModuleList([Down(nc[i], nc[i + 1]) for i in range(len(nc) - 2)])

        # padding H and W to next multiple of 8
        img_shape = psf.shape[-3:-1]
        self.top = (8 - img_shape[0] % 8) // 2
        self.bottom = (8 - img_shape[0] % 8) - self.top
        self.left = (8 - img_shape[1] % 8) // 2
        self.right = (8 - img_shape[1] % 8) - self.left

        self._psf_shape = psf.shape
        self._psf = convert_to_NCHW(psf)
        self._psf = torch.nn.functional.pad(
            self._psf, (self.left, self.right, self.top, self.bottom), mode="constant", value=0
        )
        self._n_iter = 1
        self._convolver = RealFFTConvolve2D(psf, pad=True, rgb=True if out_channels == 3 else False)

        self.set_pre_process(pre_process)
        self.skip_pre = skip_pre

    def _prepare_process_block(self, process):
        """
        Method for preparing the pre or post process block.

        Parameters
        ----------
        process : :py:class:`function` or :py:class:`~torch.nn.Module`, optional
            Pre or post process block to prepare.
        """
        if isinstance(process, torch.nn.Module):
            # If the post_process is a torch module, we assume it is a DruNet like network.
            from lensless.recon.utils import get_drunet_function_v2

            process_model = process
            process_function = get_drunet_function_v2(process_model, mode="train")
        elif process is not None:
            # Otherwise, we assume it is a function.
            assert callable(process), "pre_process must be a callable function"
            process_function = process
            process_model = None
        else:
            process_function = None
            process_model = None

        if process_function is not None:
            process_param = torch.nn.Parameter(torch.tensor([1.0], device=self._psf.device))
        else:
            process_param = None

        return process_function, process_model, process_param

    def set_pre_process(self, pre_process):
        (
            self.pre_process,
            self.pre_process_model,
            self.pre_process_param,
        ) = self._prepare_process_block(pre_process)

    def forward(self, batch, psfs=None, **kwargs):

        if psfs is None:
            psf = self._psf.to(batch.device)
        else:
            psf = convert_to_NCHW(psfs).to(batch.device)
            psf = torch.nn.functional.pad(
                psf, (self.left, self.right, self.top, self.bottom), mode="constant", value=0
            )
        n_depth = batch[0].shape[-4]
        if n_depth > 1:
            raise NotImplementedError("3D not implemented yet.")

        # pre process data
        if self.pre_process is not None and not self.skip_pre:
            device_before = batch.device
            batch = self.pre_process(batch, self.pre_process_param)
            batch = batch.to(device_before)

        # pad to multiple of 8
        batch = convert_to_NCHW(batch)
        batch = torch.nn.functional.pad(
            batch, (self.left, self.right, self.top, self.bottom), mode="constant", value=0
        )

        # -- downsample
        x_inter = [self.inc(batch)]
        for i in range(len(self.down_layers)):
            x_inter.append(self.down_layers[i](x_inter[-1]))

        # -- multi-scale Wiener filtering
        psf_multi = [self.inc0(self.w * psf)]
        for i in range(len(self.psf_down)):
            psf_multi.append(self.psf_down[i](psf_multi[-1]))
        for i in range(len(psf_multi)):
            x_inter[i] = WieNer(x_inter[i], psf_multi[i], self.delta[i])

        # upsample
        batch = self.up_layers[0](x_inter[-1], x_inter[-2])
        for i in range(len(self.up_layers) - 1):
            batch = self.up_layers[i + 1](batch, x_inter[-i - 3])
        batch = self.outc(batch)

        # back to original shape
        batch = batch[..., self.top : -self.bottom, self.left : -self.right]
        batch = convert_to_NDCHW(batch, n_depth)

        # normalize to [0,1], TODO use sigmoid instead?
        batch = (batch + 1) / 2
        batch = torch.clip(batch, min=0.0)
        return batch

    def reset(self, batch_size=1):
        # no state variables
        return

    def set_data(self, data):
        assert len(data.shape) >= 3, "Data must be at least 3D: [..., width, height, channel]."

        # assert same shapes
        assert np.all(
            self._psf_shape[-3:-1] == np.array(data.shape)[-3:-1]
        ), "PSF and data shape mismatch"

        if len(data.shape) == 3:
            self._data = data[None, None, ...]
        elif len(data.shape) == 4:
            self._data = data[None, ...]
        else:
            self._data = data

    def apply(self, **kwargs):
        # apply to data
        return self.forward(self._data, **kwargs)

    def reconstruction_error(self, prediction, lensless, **kwargs):
        convolver = self._convolver
        if not convolver.pad:
            prediction = convolver._pad(prediction)

        Fx = convolver.convolve(prediction)
        Fy = lensless

        if not convolver.pad:
            Fx = convolver._crop(Fx)

        # don't reduce batch dimension
        return torch.sum(torch.sqrt((Fx - Fy) ** 2), dim=(-1, -2, -3, -4)) / np.prod(
            prediction.shape[1:]
        )
